# coding=utf-8
from __future__ import annotations  # necessary for lazy types evaluation

__author__ = "Dimitris Karkalousos"
# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/connectors/save_restore_connector.py

import os
import shutil
import tarfile
import tempfile
import uuid
from typing import Optional, Set, Union

import torch
from omegaconf import DictConfig, OmegaConf
from omegaconf.omegaconf import open_dict
from pytorch_lightning.trainer.trainer import Trainer

# to avoid circular import do not import ModelPT directly
from atommic.core import classes as atommic_classes
from atommic.utils import logging, model_utils
from atommic.utils.app_state import AppState
from atommic.utils.get_rank import is_global_rank_zero


class SaveRestoreConnector:
    """This class is used to save and restore the model state."""

    def __init__(self) -> None:
        """Inits :class:`SaveRestoreConnector`."""
        self._model_config_yaml = "model_config.yaml"
        self._model_weights_ckpt = "model_weights.ckpt"
        self._model_extracted_dir = None

    def save_to(self, model: "atommic.ModelPT", save_path: str):  # type: ignore  # noqa: F821
        """Saves model instance (weights and configuration) into .atommic file.

        You can use "restore_from" method to fully restore instance from .atommic file.

        .atommic file is an archive (tar.gz) with the following:
        - model_config.yaml - model configuration in .yaml format. You can deserialize this into cfg argument for
        model's constructor
        - model_weights.ckpt - model checkpoint

        Parameters
        ----------
        model : ModelPT
            ModelPT object to be saved.
        save_path : str
            Path to .atommic file where model instance should be saved
        """
        if is_global_rank_zero():
            with tempfile.TemporaryDirectory() as tmpdir:
                config_yaml = os.path.join(tmpdir, self.model_config_yaml)
                model_weights = os.path.join(tmpdir, self.model_weights_ckpt)
                model.to_config_file(path2yaml_file=config_yaml)
                # update subconfigs, if there are child model, since child model can change its config
                self._update_subconfigs(model, path2yaml_file=config_yaml)
                if model.has_native_or_submodules_artifacts():
                    self._handle_artifacts(model, atommic_file_folder=tmpdir)
                    # We should not update self._cfg here - the model can still be in use
                    self._update_artifact_paths(model, path2yaml_file=config_yaml)
                self._save_state_dict_to_disk(model.state_dict(), model_weights)
                self._make_atommic_file_from_folder(filename=save_path, source_dir=tmpdir)
        else:
            return

    def load_config_and_state_dict(
        self,
        calling_cls,
        restore_path: str,
        override_config_path: Optional[Union[OmegaConf, str]] = None,
        map_location: Optional[torch.device] = None,
        strict: bool = True,  # pylint: disable=unused-argument
        return_config: bool = False,
        trainer: Trainer = None,
    ):
        """Restores model instance (weights and configuration) into .atommic file

        Parameters
        ----------
        calling_cls : class
            Class of the model to be restored.
        restore_path : str
            Path to .atommic file from which model should be instantiated
        override_config_path : Optional[Union[OmegaConf, str]]
            Path to a yaml config that will override the internal config file or an OmegaConf/DictConfig object
            representing the model config.
        map_location : Optional[torch.device]
            Optional torch.device() to map the instantiated model to a device. Default is ``None``, it will select a
            GPU if available, falling back to CPU otherwise.
        strict : bool
            Passed to load_state_dict. When set to False, the model will be able to load a checkpoint that has
            more parameters than the model itself. Default is ``True``.
        return_config : bool
            If set to true, will return just the underlying config of the restored model as an OmegaConf DictConfig
            object without instantiating the model.
        trainer : Trainer
            Optional trainer object to be used for model parallelism.

        Example
        -------
            ```
            model = atommic.collections.asr.models.EncDecCTCModel.restore_from('asr.atommic')
            assert isinstance(model, atommic.collections.asr.models.EncDecCTCModel)
            ```

        Returns
        -------
            An instance of type cls or its underlying config (if return_config is set).
        """
        # Get path where the command is executed - the artifacts will be "retrieved" there (original .atommic behavior)
        cwd = os.getcwd()

        if map_location is None:
            if torch.cuda.is_available():
                map_location = torch.device("cuda")
            else:
                map_location = torch.device("cpu")

        app_state = AppState()
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                # Check if self.model_extracted_dir is set, and is a valid path
                if self.model_extracted_dir is not None and os.path.isdir(self.model_extracted_dir):
                    # Log that atommic will use the provided `model_extracted_dir`
                    logging.info(
                        "Restoration will occur within pre-extracted directory : " f"`{self.model_extracted_dir}`."
                    )
                    # Override `tmpdir` above with the pre-extracted `model_extracted_dir`
                    tmpdir = self.model_extracted_dir
                else:
                    # Extract the atommic file into the temporary directory
                    self._unpack_atommic_file(
                        path2file=restore_path, out_folder=tmpdir, extract_config_only=return_config is True
                    )

                # Change current working directory to the temporary directory
                os.chdir(tmpdir)
                if override_config_path is None:
                    config_yaml = self.model_config_yaml
                else:
                    # can be str path or OmegaConf / DictConfig object
                    config_yaml = override_config_path
                if not isinstance(config_yaml, (OmegaConf, DictConfig)):
                    conf = OmegaConf.load(config_yaml)
                else:
                    conf = config_yaml
                    if override_config_path is not None:
                        # Resolve the override config
                        conf = OmegaConf.to_container(conf, resolve=True)
                        conf = OmegaConf.create(conf)
                # If override is top level config, extract just `model` from it
                if "model" in conf:
                    conf = conf.model

                if return_config:
                    instance = conf
                    return instance

                if app_state.model_parallel_rank is not None and app_state.model_parallel_size > 1:
                    model_weights = self._inject_model_parallel_rank_for_ckpt(tmpdir, self.model_weights_ckpt)
                else:
                    model_weights = os.path.join(tmpdir, self.model_weights_ckpt)
                OmegaConf.set_struct(conf, True)
                os.chdir(cwd)
                # get the class
                calling_cls._set_model_restore_state(  # pylint: disable=protected-access
                    is_being_restored=True, folder=tmpdir
                )
                instance = calling_cls.from_config_dict(config=conf, trainer=trainer)
                instance = instance.to(map_location)
                # add load_state_dict override
                if app_state.model_parallel_size is not None and app_state.model_parallel_size > 1:
                    model_weights = self._inject_model_parallel_rank_for_ckpt(tmpdir, self.model_weights_ckpt)
                state_dict = self._load_state_dict_from_disk(model_weights, map_location=map_location)
            finally:
                os.chdir(cwd)

        return (conf, instance, state_dict)

    @staticmethod
    def load_instance_with_state_dict(instance, state_dict, strict):
        """Loads the state dict into the instance."""
        instance.load_state_dict(state_dict, strict=strict)
        instance._set_model_restore_state(is_being_restored=False)  # pylint: disable=protected-access

    def restore_from(
        self,
        calling_cls,
        restore_path: str,
        override_config_path: Optional[Union[OmegaConf, str]] = None,
        map_location: Optional[torch.device] = None,
        strict: bool = True,
        return_config: bool = False,
        trainer: Trainer = None,
    ):
        """Restores model instance (weights and configuration) into .atommic file

        Parameters
        ----------
        calling_cls : class
            The class of the model to be restored.
        restore_path : str
            Path to .atommic file from which model should be instantiated.
        override_config_path : str or OmegaConf/DictConfig object, optional
            Path to a yaml config that will override the internal config file or an OmegaConf/DictConfig object
            representing the model config.
        map_location : torch.device, optional
            Optional torch.device() to map the instantiated model to a device. By default (None), it will select a
            GPU if available, falling back to CPU otherwise.
        strict : bool, optional
            Passed to load_state_dict. Default is ``True``.
        return_config : bool, optional
            If set to true, will return just the underlying config of the restored model as an OmegaConf/DictConfig
            object without instantiating the model.
        trainer : Trainer, optional
            Optional trainer object to be used for restoring the model.

        Returns
        -------
        An instance of type cls or its underlying config (if return_config is set).
        """
        # Get path where the command is executed - the artifacts will be "retrieved" there (original .atommic behavior)
        loaded_params = self.load_config_and_state_dict(
            calling_cls, restore_path, override_config_path, map_location, strict, return_config, trainer
        )
        if not isinstance(loaded_params, tuple) or return_config is True:
            return loaded_params
        _, instance, state_dict = loaded_params
        self.load_instance_with_state_dict(instance, state_dict, strict)
        logging.info(f"Model {instance.__class__.__name__} was successfully restored from {restore_path}.")
        return instance, state_dict

    def extract_state_dict_from(self, restore_path: str, save_dir: str, split_by_module: bool = False):
        """Extract the state dict(s) from a provided .atommic tarfile and save it to a directory.

        Parameters
        ----------
        restore_path : str
            Path to .atommic file from which state dict(s) should be extracted.
        save_dir : str
            Directory in which the saved state dict(s) should be stored.
        split_by_module : bool, optional
            Bool flag, which determines whether the output checkpoint should be for the entire Model, or
            the individual module's that comprise the Model. Default is ``False``.

        Example
        -------
        To convert the .atommic tarfile into a single Model level PyTorch checkpoint
        ::
        state_dict = atommic.collections.asr.models.EncDecCTCModel.extract_state_dict_from('asr.atommic',
        './asr_ckpts')
        To restore a model from a Model level checkpoint
        ::
        model = atommic.collections.asr.models.EncDecCTCModel(cfg)  # or any other method of restoration
        model.load_state_dict(torch.load("./asr_ckpts/model_weights.ckpt"))
        To convert the .atommic tarfile into multiple Module level PyTorch checkpoints
        ::
        state_dict = atommic.collections.asr.models.EncDecCTCModel.extract_state_dict_from('asr.atommic',
        './asr_ckpts', split_by_module=True). To restore a module from a Module level checkpoint
        ::
        model = atommic.collections.asr.models.EncDecCTCModel(cfg)  # or any other method of restoration
        # load the individual components
        model.preprocessor.load_state_dict(torch.load("./asr_ckpts/preprocessor.ckpt"))
        model.encoder.load_state_dict(torch.load("./asr_ckpts/encoder.ckpt"))
        model.decoder.load_state_dict(torch.load("./asr_ckpts/decoder.ckpt"))

        Returns
        -------
        The state dict that was loaded from the original .atommic checkpoint.
        """
        cwd = os.getcwd()

        save_dir = os.path.abspath(save_dir)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir, exist_ok=True)

        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                self._unpack_atommic_file(path2file=restore_path, out_folder=tmpdir)
                os.chdir(tmpdir)
                model_weights = os.path.join(tmpdir, self.model_weights_ckpt)
                state_dict = self._load_state_dict_from_disk(model_weights)

                if not split_by_module:
                    filepath = os.path.join(save_dir, self.model_weights_ckpt)
                    self._save_state_dict_to_disk(state_dict, filepath)

                else:
                    key_set = {key.split(".")[0] for key in state_dict.keys()}
                    for primary_key in key_set:
                        inner_keys = [key for key in state_dict.keys() if key.split(".")[0] == primary_key]
                        state_dict_subset = {
                            ".".join(inner_key.split(".")[1:]): state_dict[inner_key] for inner_key in inner_keys
                        }
                        filepath = os.path.join(save_dir, f"{primary_key}.ckpt")
                        self._save_state_dict_to_disk(state_dict_subset, filepath)

                logging.info(f"Checkpoints from {restore_path} were successfully extracted into {save_dir}.")
            finally:
                os.chdir(cwd)

        return state_dict

    @staticmethod
    def register_artifact(model, config_path: str, src: str, verify_src_exists: bool = True):
        """Register model artifacts with this function. These artifacts (files) will be included inside .atommic file
        when model.save_to("mymodel.atommic") is called.

        How it works:
        1. It always returns existing absolute path which can be used during Model constructor call. EXCEPTION: src is
        None or "" in which case nothing will be done and src will be returned
        2. It will add (config_path, model_utils.ArtifactItem()) pair to self.artifacts. If "src" is local existing
        path, then it will be returned in absolute path form. elif "src" starts with
        "atommic_file:unique_artifact_name": .atommic will be untarred to a temporary folder location and an actual
        existing path will be returned else an error will be raised.

            .. code-block::

              If "src" is local existing path:
                  then it will be returned in absolute path form
              elif "src" starts with "atommic_file:unique_artifact_name":
                  .atommic will be untarred to a temporary folder location and an actual existing path will be returned
              else:
                  an error will be raised.

        WARNING: use .register_artifact calls in your models' constructors.
        The returned path is not guaranteed to exist after you have exited your model's constructor.

        Parameters
        ----------
        model : ModelPT
            ModelPT object to register artifact for.
        config_path : str
            Artifact key. Usually corresponds to the model config.
        src : str
            Path to artifact.
        verify_src_exists : bool, optional
            If set to False, then the artifact is optional and register_artifact will return None even if src is not
            found. Default is ``True``.

        Returns
        --------
        If src is not None or empty it always returns absolute path which is guaranteed to exist during model instance
         life.
        """
        app_state = AppState()

        artifact_item = model_utils.ArtifactItem()

        # This is for backward compatibility, if the src objects exists simply inside the tarfile
        # without its key having been overridden, this pathway will be used.
        src_obj_name = os.path.basename(src)
        if app_state.atommic_file_folder is not None:
            src_obj_path = os.path.abspath(os.path.join(app_state.atommic_file_folder, src_obj_name))
        else:
            src_obj_path = src_obj_name

        # src is a local existing path - register artifact and return exact same path for usage by the model
        if os.path.exists(os.path.abspath(src)):
            return_path = os.path.abspath(src)
            artifact_item.path_type = model_utils.ArtifactPathType.LOCAL_PATH

        elif src.startswith("atommic:"):
            return_path = os.path.abspath(os.path.join(app_state.atommic_file_folder, src[5:]))
            artifact_item.path_type = model_utils.ArtifactPathType.TAR_PATH

        elif os.path.exists(src_obj_path):
            return_path = src_obj_path
            artifact_item.path_type = model_utils.ArtifactPathType.TAR_PATH
        elif verify_src_exists:
            raise FileNotFoundError(
                f"src path does not exist or it is not a path in atommic file. src value I got was: {src}. "
                f"Absolute: {os.path.abspath(src)}"
            )
        else:
            # artifact is optional and we simply return None
            return None

        if not os.path.exists(return_path):
            raise AssertionError

        artifact_item.path = os.path.abspath(src)
        model.artifacts[config_path] = artifact_item
        # we were called by ModelPT
        if hasattr(model, "cfg"):
            with open_dict(model._cfg):  # pylint: disable=protected-access
                OmegaConf.update(model.cfg, config_path, return_path)
        return return_path

    def _handle_artifacts(self, model, atommic_file_folder):  # noqa: MC0001
        """This method is called by ModelPT.save_to() and ModelPT.load_from(). It will handle all artifacts and save
        them to the atommic file.

        Parameters
        ----------
        model : ModelPT
            ModelPT object to handle artifacts for.
        atommic_file_folder : str
            Path to temporary folder where atommic file will be untarred.
        """
        tarfile_artifacts = []
        app_state = AppState()

        # aggregate artifacts from self and all children recursively
        artifacts_containers = []
        for _, config_path, module in model.named_atommic_modules():
            if module.has_artifacts():  # atommic model with artifacts
                artifacts_containers.append((config_path, module.artifacts))

        if len(artifacts_containers) > 0 and (not hasattr(model, "artifacts") or model.artifacts is None):
            # model has no artifacts, but submodules have some
            model.artifacts = {}

        for config_path, artifacts in artifacts_containers:
            for subconf_path, artiitem in artifacts.items():
                conf_path = f"{config_path}.{subconf_path}" if config_path else f"{subconf_path}"
                if artiitem.path_type == model_utils.ArtifactPathType.LOCAL_PATH:
                    if not os.path.exists(artiitem.path):
                        raise FileNotFoundError(f"Artifact {conf_path} not found at location: {artiitem.path}")

                    # Generate new uniq artifact name and copy it to atommic_file_folder
                    # Note uuid.uuid4().hex is guaranteed to be 32 character long
                    artifact_base_name = os.path.basename(artiitem.path)
                    artifact_uniq_name = f"{uuid.uuid4().hex}_{artifact_base_name}"
                    shutil.copy2(artiitem.path, os.path.join(atommic_file_folder, artifact_uniq_name))

                    # Update artifacts registry
                    artiitem.hashed_path = "atommic:" + artifact_uniq_name
                    model.artifacts[conf_path] = artiitem

                elif artiitem.path_type == model_utils.ArtifactPathType.TAR_PATH:
                    # process all tarfile artifacts in one go, so preserve key-value pair
                    tarfile_artifacts.append((conf_path, artiitem))
                    if subconf_path:  # artifact from submodule
                        model.artifacts[conf_path] = artiitem

                else:
                    raise ValueError("Directly referencing artifacts from other atommic files isn't supported yet")

        # Process current tarfile artifacts by unpacking the previous tarfile and extract the artifacts
        # that are currently required.
        # artifacts can be native (from the model itself) and from submodules
        # model + submodules restoration paths, handle only unique paths
        restoration_paths: Set[str] = set()
        model_metadata = app_state.get_model_metadata_from_guid(model.model_guid)
        if model_metadata.restoration_path is not None:
            restoration_paths.add(model_metadata.restoration_path)
        # aggregate restoration paths for all submodules recursively
        for module in model.modules():
            if isinstance(module, atommic_classes.modelPT.ModelPT):  # if atommic model
                submodule_restoration_path = app_state.get_model_metadata_from_guid(module.model_guid).restoration_path
                if submodule_restoration_path is not None:
                    restoration_paths.add(submodule_restoration_path)
        if len(tarfile_artifacts) > 0 and len(restoration_paths) == 0:
            # TODO: see cases when this can occur, and if we can fix them
            logging.warning("Model contains registered artifacts, but no restoration paths found")
        if len(tarfile_artifacts) > 0 and len(restoration_paths) > 0:
            # Need to step into atommic archive to extract file
            # Get path where the command is executed - the artifacts will be "retrieved" there
            # (original .atommic behavior)
            cwd = os.getcwd()
            with tempfile.TemporaryDirectory() as archive_dir:
                # Step into the atommic archive to try and find the file
                try:
                    # unpack all restorations paths (atommic checkpoints)
                    # in atommic checkpoints all resources contain hash in name, so there should be no collisions
                    for path in restoration_paths:
                        if self.model_extracted_dir:
                            shutil.copytree(src=path, dst=archive_dir, dirs_exist_ok=True)
                        else:
                            self._unpack_atommic_file(path2file=path, out_folder=archive_dir)
                    os.chdir(archive_dir)
                    for conf_path, artiitem in tarfile_artifacts:
                        # Get basename and copy it to atommic_file_folder
                        if "atommic:" in artiitem.path:
                            artifact_base_name = artiitem.path.split("atommic:")[1]
                        else:
                            artifact_base_name = os.path.basename(artiitem.path)
                        # no need to hash here as we are in tarfile_artifacts which are already hashed
                        artifact_uniq_name = artifact_base_name
                        shutil.copy2(artifact_base_name, os.path.join(atommic_file_folder, artifact_uniq_name))

                        # Update artifacts registry
                        new_artiitem = model_utils.ArtifactItem()
                        new_artiitem.path = "atommic:" + artifact_uniq_name
                        new_artiitem.path_type = model_utils.ArtifactPathType.TAR_PATH
                        model.artifacts[conf_path] = new_artiitem
                finally:
                    # change back working directory
                    os.chdir(cwd)

    @staticmethod
    def _update_subconfigs(model: "atommic_classes.ModelPT", path2yaml_file):  # type: ignore
        """Update subconfigs if ModelPT has submodules. Should be called before updating artifacts paths."""
        if not model.has_atommic_submodules():
            # no submodules => nothing to update
            return
        conf = OmegaConf.load(path2yaml_file)
        # update subconfigs for all children recursively, parent configs updated before children
        for _, conf_path, submodule in model.named_atommic_modules():
            if not conf_path:  # self
                continue
            OmegaConf.update(conf, conf_path, submodule.cfg)
        with open(path2yaml_file, "w", encoding="utf-8") as fout:
            OmegaConf.save(config=conf, f=fout, resolve=True)

    @staticmethod
    def _update_artifact_paths(model, path2yaml_file):
        """This method is called by ModelPT.save_to() and ModelPT.load_from() to update the artifact paths in the
        model.
        """
        if hasattr(model, "artifacts") and model.artifacts is not None and len(model.artifacts) > 0:
            conf = OmegaConf.load(path2yaml_file)
            for conf_path, item in model.artifacts.items():
                if item.hashed_path is None:
                    OmegaConf.update(conf, conf_path, item.path)
                else:
                    OmegaConf.update(conf, conf_path, item.hashed_path)
            with open(path2yaml_file, "w", encoding="utf-8") as fout:
                OmegaConf.save(config=conf, f=fout, resolve=True)

    @staticmethod
    def _inject_model_parallel_rank_for_ckpt(dirname, basename):
        """This method is called by ModelPT.save_to() and ModelPT.load_from() to inject the parallel rank of the
        process into the checkpoint file name.
        """
        model_weights = os.path.join(dirname, basename)
        model_weights = model_utils.inject_model_parallel_rank(model_weights)
        return model_weights

    @staticmethod
    def _make_atommic_file_from_folder(filename, source_dir):
        """The method is called by ModelPT.save_to() and ModelPT.load_from() to create a atommic file from a folder."""
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)
        with tarfile.open(filename, "w") as tar:
            tar.add(source_dir, arcname=".")

    @staticmethod
    def _unpack_atommic_file(path2file: str, out_folder: str, extract_config_only: bool = False) -> str:
        """This method is called by ModelPT.save_to() and ModelPT.load_from() to unpack a atommic file."""
        if not os.path.exists(path2file):
            raise FileNotFoundError(f"{path2file} does not exist")
        # we start with an assumption of uncompressed tar, which should be true for versions 1.7.0 and above
        tar_header = "r:"
        try:
            tar_test = tarfile.open(path2file, tar_header)  # pylint: disable=consider-using-with
            tar_test.close()
        except tarfile.ReadError:
            # can be older checkpoint => try compressed tar
            tar_header = "r:gz"
        tar = tarfile.open(path2file, tar_header)  # pylint: disable=consider-using-with
        if not extract_config_only:
            tar.extractall(path=out_folder)
        else:
            members = [x for x in tar.getmembers() if ".yaml" in x.name]
            tar.extractall(path=out_folder, members=members)
        tar.close()
        return out_folder

    @staticmethod
    def _save_state_dict_to_disk(state_dict, filepath):
        """This method is called by ModelPT.save_to() and ModelPT.load_from() to save the state dict to disk."""
        torch.save(state_dict, filepath)

    @staticmethod
    def _load_state_dict_from_disk(model_weights, map_location="cpu"):
        """This method is called by ModelPT.save_to() and ModelPT.load_from() to load the state dict from disk."""
        return torch.load(model_weights, map_location=map_location)

    @property
    def model_config_yaml(self) -> str:
        """This property is used to get the path to the model config yaml file."""
        return self._model_config_yaml

    @model_config_yaml.setter
    def model_config_yaml(self, path: str):
        """This property is used to set the path to the model config yaml file."""
        self._model_config_yaml = path

    @property
    def model_weights_ckpt(self) -> str:
        """This property is used to get the path to the model weights ckpt file."""
        return self._model_weights_ckpt

    @model_weights_ckpt.setter
    def model_weights_ckpt(self, path: str):
        """This property is used to set the path to the model weights ckpt file."""
        self._model_weights_ckpt = path

    @property
    def model_extracted_dir(self) -> Optional[str]:
        return self._model_extracted_dir

    @model_extracted_dir.setter
    def model_extracted_dir(self, path: None):
        self._model_extracted_dir = path



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from skimage.restoration import unwrap_phase
from torch.nn import functional as F

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.transforms import (
    N2R,
    SSDU,
    Composer,
    Cropper,
    EstimateCoilSensitivityMaps,
    GeometricDecompositionCoilCompression,
    Masker,
    NoisePreWhitening,
    Normalizer,
    ZeroFillingPadding,
)
from atommic.collections.common.parts.utils import add_coil_dim_if_singlecoil
from atommic.collections.common.parts.utils import coil_combination_method as coil_combination_method_func
from atommic.collections.common.parts.utils import is_none, sense, to_tensor
from atommic.collections.motioncorrection.parts.motionsimulation import MotionSimulation

__all__ = ["qMRIDataTransforms"]


class qMRIDataTransforms:
    """Data transforms for quantitative MRI.

    Returns
    -------
    qMRIDataTransforms
        Preprocessed data for quantitative MRI.
    """

    def __init__(
        self,
        TEs: Optional[List[float]],
        precompute_quantitative_maps: bool = True,
        qmaps_scaling_factor: float = 1.0,
        dataset_format: str = None,
        apply_prewhitening: bool = False,
        find_patch_size: bool = True,
        prewhitening_scale_factor: float = 1.0,
        prewhitening_patch_start: int = 10,
        prewhitening_patch_length: int = 30,
        apply_gcc: bool = False,
        gcc_virtual_coils: int = 10,
        gcc_calib_lines: int = 24,
        gcc_align_data: bool = True,
        apply_random_motion: bool = False,
        random_motion_type: str = "gaussian",
        random_motion_percentage: Sequence[int] = (10, 10),
        random_motion_angle: int = 10,
        random_motion_translation: int = 10,
        random_motion_center_percentage: float = 0.02,
        random_motion_num_segments: int = 8,
        random_motion_random_num_segments: bool = True,
        random_motion_non_uniform: bool = False,
        estimate_coil_sensitivity_maps: bool = False,
        coil_sensitivity_maps_type: str = "ESPIRiT",
        coil_sensitivity_maps_gaussian_sigma: float = 0.0,
        coil_sensitivity_maps_espirit_threshold: float = 0.05,
        coil_sensitivity_maps_espirit_kernel_size: int = 6,
        coil_sensitivity_maps_espirit_crop: float = 0.95,
        coil_sensitivity_maps_espirit_max_iters: int = 30,
        coil_combination_method: str = "SENSE",
        dimensionality: int = 2,
        mask_func: Optional[List] = None,
        shift_mask: bool = False,
        mask_center_scale: Optional[float] = 0.02,
        partial_fourier_percentage: float = 0.0,
        remask: bool = False,
        ssdu: bool = False,
        ssdu_mask_type: str = "Gaussian",
        ssdu_rho: float = 0.4,
        ssdu_acs_block_size: Sequence[int] = (4, 4),
        ssdu_gaussian_std_scaling_factor: float = 4.0,
        ssdu_outer_kspace_fraction: float = 0.0,
        ssdu_export_and_reuse_masks: bool = False,
        n2r: bool = False,
        n2r_supervised_rate: float = 0.0,
        n2r_probability: float = 0.0,
        n2r_std_devs: Tuple[float, float] = None,
        n2r_rhos: Tuple[float, float] = None,
        n2r_use_mask: bool = False,
        unsupervised_masked_target: bool = False,
        crop_size: Optional[Tuple[int, int]] = None,
        kspace_crop: bool = False,
        crop_before_masking: bool = True,
        kspace_zero_filling_size: Optional[Tuple] = None,
        normalize_inputs: bool = True,
        normalization_type: str = "max",
        kspace_normalization: bool = False,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 0,
        consecutive_slices: int = 1,  # pylint: disable=unused-argument
        use_seed: bool = True,
    ):
        """Inits :class:`qMRIDataTransforms`.

        Parameters
        ----------
        TEs : Optional[List[float]]
            Echo times.
        precompute_quantitative_maps : bool, optional
            Precompute quantitative maps. Default is ``True``.
        qmaps_scaling_factor : float, optional
            Quantitative maps scaling factor. Default is ``1e-3``.
        dataset_format : str, optional
            The format of the dataset. For example, ``'custom_dataset'`` or ``'public_dataset_name'``.
            Default is ``None``.
        apply_prewhitening : bool, optional
            Apply prewhitening. If ``True`` then the prewhitening arguments are used. Default is ``False``.
        find_patch_size : bool, optional
            Find optimal patch size (automatically) to calculate psi. If False, patch_size must be defined.
            Default is ``True``.
        prewhitening_scale_factor : float, optional
            Prewhitening scale factor. Default is ``1.0``.
        prewhitening_patch_start : int, optional
            Prewhitening patch start. Default is ``10``.
        prewhitening_patch_length : int, optional
            Prewhitening patch length. Default is ``30``.
        apply_gcc : bool, optional
            Apply Geometric Decomposition Coil Compression. If ``True`` then the GCC arguments are used.
            Default is ``False``.
        gcc_virtual_coils : int, optional
            GCC virtual coils. Default is ``10``.
        gcc_calib_lines : int, optional
            GCC calibration lines. Default is ``24``.
        gcc_align_data : bool, optional
            GCC align data. Default is ``True``.
        apply_random_motion : bool, optional
            Simulate random motion in k-space. Default is ``False``.
        random_motion_type : str, optional
            Random motion type. It can be one of the following: ``piecewise_transient``, ``piecewise_constant``,
            ``gaussian``. Default is ``gaussian``.
        random_motion_percentage : Sequence[int], optional
            Random motion percentage. For example, 10%-20% motion can be defined as ``(10, 20)``.
            Default is ``(10, 10)``.
        random_motion_angle : float, optional
            Random motion angle. Default is ``10.0``.
        random_motion_translation : float, optional
            Random motion translation. Default is ``10.0``.
        random_motion_center_percentage : float, optional
            Random motion center percentage. Default is ``0.0``.
        random_motion_num_segments : int, optional
            Random motion number of segments to partition the k-space. Default is ``8``.
        random_motion_random_num_segments : bool, optional
            Whether to randomly generate the number of segments. Default is ``True``.
        random_motion_non_uniform : bool, optional
            Random motion non-uniform sampling. Default is ``False``.
        estimate_coil_sensitivity_maps : bool, optional
            Automatically estimate coil sensitivity maps. Default is ``False``. If ``True`` then the coil sensitivity
            maps arguments are used. Note that this is different from the ``estimate_coil_sensitivity_maps_with_nn``
            argument, which uses a neural network to estimate the coil sensitivity maps. The
            ``estimate_coil_sensitivity_maps`` estimates the coil sensitivity maps with methods such as ``ESPIRiT``,
            ``RSS`` or ``UNit``. ``ESPIRiT`` is the ``Eigenvalue to Self-Consistent Parallel Imaging Reconstruction
            Technique`` method. ``RSS`` is the ``Root Sum of Squares``  method. ``UNit`` returns a uniform coil
            sensitivity map.
        coil_sensitivity_maps_type : str, optional
            Coil sensitivity maps type. It can be one of the following: ``ESPIRiT``, ``RSS`` or ``UNit``. Default is
            ``ESPIRiT``.
        coil_sensitivity_maps_gaussian_sigma : float, optional
            Coil sensitivity maps Gaussian sigma. Default is ``0.0``.
        coil_sensitivity_maps_espirit_threshold : float, optional
            Coil sensitivity maps ESPRIT threshold. Default is ``0.05``.
        coil_sensitivity_maps_espirit_kernel_size : int, optional
            Coil sensitivity maps ESPRIT kernel size. Default is ``6``.
        coil_sensitivity_maps_espirit_crop : float, optional
            Coil sensitivity maps ESPRIT crop. Default is ``0.95``.
        coil_sensitivity_maps_espirit_max_iters : int, optional
            Coil sensitivity maps ESPRIT max iterations. Default is ``30``.
        coil_combination_method : str, optional
            Coil combination method. Default is ``"SENSE"``.
        dimensionality : int, optional
            Dimensionality. Default is ``2``.
        mask_func : Optional[List["MaskFunc"]], optional
            Mask function to retrospectively undersample the k-space. Default is ``None``.
        shift_mask : bool, optional
            Whether to shift the mask. This needs to be set alongside with the ``fft_centered`` argument.
            Default is ``False``.
        mask_center_scale : Optional[float], optional
            Center scale of the mask. This defines how much densely sampled will be the center of k-space.
            Default is ``0.02``.
        partial_fourier_percentage : float, optional
            Whether to simulate a half scan. Default is ``0.0``.
        remask : bool, optional
            Use the same mask. Default is ``False``.
        ssdu : bool, optional
            Whether to apply Self-Supervised Data Undersampling (SSDU) masks. Default is ``False``.
        ssdu_mask_type: str, optional
            Mask type. It can be one of the following:
            - "Gaussian": Gaussian sampling.
            - "Uniform": Uniform sampling.
            Default is "Gaussian".
        ssdu_rho: float, optional
            Split ratio for training and loss masks. Default is ``0.4``.
        ssdu_acs_block_size: tuple, optional
            Keeps a small acs region fully-sampled for training masks, if there is no acs region. The small acs block
            should be set to zero. Default is ``(4, 4)``.
        ssdu_gaussian_std_scaling_factor: float, optional
            Scaling factor for standard deviation of the Gaussian noise. If Uniform is select this factor is ignored.
            Default is ``4.0``.
        ssdu_outer_kspace_fraction: float, optional
            Fraction of the outer k-space to be kept/unmasked. Default is ``0.0``.
        ssdu_export_and_reuse_masks: bool, optional
            Whether to export and reuse the masks. Default is ``False``.
        n2r : bool, optional
            Whether to apply Noise to Reconstruction (N2R) masks. Default is ``False``.
        n2r_supervised_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the subjects should be loaded for Noise to
            Reconstruction (N2R) supervised loss, if N2R is enabled. Default is ``0.0``.
        n2r_probability : float, optional
            Probability of applying N2R. Default is ``0.0``.
        n2r_std_devs : Tuple[float, float], optional
            Standard deviations for the noise. Default is ``(0.0, 0.0)``.
        n2r_rhos : Tuple[float, float], optional
            Rho values for the noise. Default is ``(0.0, 0.0)``.
        n2r_use_mask : bool, optional
            Whether to use a mask for N2R. Default is ``False``.
        unsupervised_masked_target : bool, optional
            Whether to use the masked initial estimation for unsupervised learning. Default is ``False``.
        crop_size : Optional[Tuple[int, int]], optional
            Center crop size. It applies cropping in image space. Default is ``None``.
        kspace_crop : bool, optional
            Whether to crop in k-space. Default is ``False``.
        crop_before_masking : bool, optional
            Whether to crop before masking. Default is ``True``.
        kspace_zero_filling_size : Optional[Tuple], optional
            Whether to apply zero filling in k-space. Default is ``None``.
        normalize_inputs : bool, optional
            Whether to normalize the inputs. Default is ``True``.
        normalization_type : str, optional
            Normalization type. Can be ``max`` or ``mean`` or ``minmax``. Default is ``max``.
        kspace_normalization : bool, optional
            Whether to normalize the k-space. Default is ``False``.
        fft_centered : bool, optional
            Whether to center the FFT. Default is ``False``.
        fft_normalization : str, optional
            FFT normalization. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``0``, meaning that the coil dimension is the first dimension before applying
            batch.
        consecutive_slices : int, optional
            Consecutive slices. Default is ``1``.
        use_seed : bool, optional
            Whether to use seed. Default is ``True``.
        """
        super().__init__()

        if not precompute_quantitative_maps:
            raise ValueError(
                "Loading quantitative maps from disk is not supported yet. "
                "Please set precompute_quantitative_maps to True."
            )
        self.precompute_quantitative_maps = precompute_quantitative_maps

        if TEs is None:
            raise ValueError("Please specify echo times (TEs).")
        self.TEs = TEs
        self.qmaps_scaling_factor = qmaps_scaling_factor

        self.dataset_format = dataset_format

        self.coil_combination_method = coil_combination_method
        self.kspace_crop = kspace_crop
        self.crop_before_masking = crop_before_masking

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim - 1 if dimensionality == 2 else coil_dim

        self.prewhitening = (
            NoisePreWhitening(
                find_patch_size=find_patch_size,
                patch_size=[
                    prewhitening_patch_start,
                    prewhitening_patch_length + prewhitening_patch_start,
                    prewhitening_patch_start,
                    prewhitening_patch_length + prewhitening_patch_start,
                ],
                scale_factor=prewhitening_scale_factor,
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            if apply_prewhitening
            else None
        )

        self.gcc = (
            GeometricDecompositionCoilCompression(
                virtual_coils=gcc_virtual_coils,
                calib_lines=gcc_calib_lines,
                align_data=gcc_align_data,
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            if apply_gcc
            else None
        )

        self.random_motion = (
            MotionSimulation(
                motion_type=random_motion_type,
                angle=random_motion_angle,
                translation=random_motion_translation,
                center_percentage=random_motion_center_percentage,
                motion_percentage=random_motion_percentage,
                num_segments=random_motion_num_segments,
                random_num_segments=random_motion_random_num_segments,
                non_uniform=random_motion_non_uniform,
                spatial_dims=self.spatial_dims,
            )
            if apply_random_motion
            else None
        )

        self.coil_sensitivity_maps_estimator = (
            EstimateCoilSensitivityMaps(
                coil_sensitivity_maps_type=coil_sensitivity_maps_type.lower(),
                gaussian_sigma=coil_sensitivity_maps_gaussian_sigma,
                espirit_threshold=coil_sensitivity_maps_espirit_threshold,
                espirit_kernel_size=coil_sensitivity_maps_espirit_kernel_size,
                espirit_crop=coil_sensitivity_maps_espirit_crop,
                espirit_max_iters=coil_sensitivity_maps_espirit_max_iters,
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
                coil_dim=self.coil_dim,
            )
            if estimate_coil_sensitivity_maps
            else None
        )

        self.kspace_zero_filling = (
            ZeroFillingPadding(
                zero_filling_size=kspace_zero_filling_size,  # type: ignore
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            if not is_none(kspace_zero_filling_size)
            else None
        )

        self.shift_mask = shift_mask
        self.masking = Masker(
            mask_func=mask_func,  # type: ignore
            spatial_dims=self.spatial_dims,
            shift_mask=self.shift_mask,
            partial_fourier_percentage=partial_fourier_percentage,
            center_scale=mask_center_scale,  # type: ignore
            dimensionality=dimensionality,
            remask=remask,
            dataset_format=self.dataset_format,
        )

        self.ssdu = ssdu
        self.ssdu_masking = (
            SSDU(
                mask_type=ssdu_mask_type,
                rho=ssdu_rho,
                acs_block_size=ssdu_acs_block_size,
                gaussian_std_scaling_factor=ssdu_gaussian_std_scaling_factor,
                outer_kspace_fraction=ssdu_outer_kspace_fraction,
                export_and_reuse_masks=ssdu_export_and_reuse_masks,
            )
            if self.ssdu
            else None
        )

        self.n2r = n2r
        self.n2r_supervised_rate = n2r_supervised_rate
        self.n2r_masking = (
            N2R(
                probability=n2r_probability,
                std_devs=n2r_std_devs,  # type: ignore
                rhos=n2r_rhos,  # type: ignore
                use_mask=n2r_use_mask,
            )
            if self.n2r
            else None
        )

        self.unsupervised_masked_target = unsupervised_masked_target

        self.cropping = (
            Cropper(
                cropping_size=crop_size,  # type: ignore
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            if not is_none(crop_size)
            else None
        )

        self.normalization_type = normalization_type
        self.normalization = (
            Normalizer(
                normalization_type=self.normalization_type,
                kspace_normalization=kspace_normalization,
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            if normalize_inputs
            else None
        )

        self.prewhitening = Composer([self.prewhitening])  # type: ignore
        self.coils_shape_transforms = Composer(
            [
                self.gcc,  # type: ignore
                self.kspace_zero_filling,  # type: ignore
            ]
        )
        self.crop_normalize = Composer(
            [
                self.cropping,  # type: ignore
                self.normalization,  # type: ignore
            ]
        )
        self.cropping = Composer([self.cropping])  # type: ignore
        self.random_motion = Composer([self.random_motion])  # type: ignore
        self.normalization = Composer([self.normalization])  # type: ignore

        self.use_seed = use_seed

    def __call__(
        self,
        kspace: np.ndarray,
        sensitivity_map: np.ndarray,
        mask: np.ndarray,
        initial_prediction: np.ndarray,
        target: np.ndarray,
        attrs: Dict,
        fname: str,
        slice_idx: int,
    ) -> Tuple[
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        torch.Tensor,
        Union[List[torch.Tensor], torch.Tensor],
        Union[List[torch.Tensor], torch.Tensor],
        str,
        int,
        List[Union[float, torch.Tensor, Any]],
        Dict,
    ]:
        """Calls :class:`qMRIDataTransforms`.

        Parameters
        ----------
        kspace : np.ndarray
            The fully-sampled kspace, if exists. Otherwise, the subsampled kspace.
        sensitivity_map : np.ndarray
            The coil sensitivity map.
        mask : np.ndarray
            The subsampling mask, if exists, meaning that the data are either prospectively undersampled or the mask is
            stored and loaded. It can be a list of masks, with the subsampling, the brain, and the head mask.
        initial_prediction : np.ndarray
            The initial prediction, if exists. Otherwise, it will be estimated with the chosen coil combination method.
        target : np.ndarray
            The target, if exists. Otherwise, it will be estimated with the chosen coil combination method.
        attrs : Dict
            The attributes, if stored in the data.
        fname : str
            The file name.
        slice_idx : int
            The slice index.

        Returns
        -------
        The transformed data.
        """
        mask, anatomy_mask = mask

        if mask.ndim <= 1:
            mask = None

        kspace, masked_kspace, mask, kspace_pre_normalization_vars, acc = self.__process_kspace__(  # type: ignore
            kspace, mask, attrs, fname
        )
        sensitivity_map, sensitivity_pre_normalization_vars = self.__process_coil_sensitivities_map__(
            sensitivity_map, kspace
        )

        if self.n2r and len(masked_kspace) > 1:  # type: ignore
            prediction, prediction_pre_normalization_vars = self.__initialize_prediction__(
                initial_prediction, masked_kspace[0], sensitivity_map  # type: ignore
            )
            if isinstance(masked_kspace, list) and not masked_kspace[1][0].dim() < 2:  # type: ignore
                noise_prediction, noise_prediction_pre_normalization_vars = self.__initialize_prediction__(
                    None, masked_kspace[1], sensitivity_map  # type: ignore
                )
            else:
                noise_prediction = torch.tensor([])
                noise_prediction_pre_normalization_vars = None
            prediction = [prediction, noise_prediction]
        else:
            prediction, prediction_pre_normalization_vars = self.__initialize_prediction__(
                initial_prediction, masked_kspace, sensitivity_map  # type: ignore
            )
            noise_prediction_pre_normalization_vars = None

        if self.unsupervised_masked_target:
            target, target_pre_normalization_vars = prediction, prediction_pre_normalization_vars
        else:
            target, target_pre_normalization_vars = self.__initialize_prediction__(
                None if self.ssdu else target, kspace, sensitivity_map
            )

        if anatomy_mask.ndim != 0:
            anatomy_mask = self.cropping(torch.from_numpy(anatomy_mask))  # type: ignore

        (
            R2star_map_target,
            R2star_map_target_pre_normalization_vars,
            S0_map_target,
            S0_map_target_pre_normalization_vars,
            B0_map_target,
            B0_map_target_pre_normalization_vars,
            phi_map_target,
            phi_map_target_pre_normalization_vars,
        ) = self.__compute_quantitative_maps__(
            kspace, sensitivity_map, None, anatomy_mask
        )  # type: ignore
        (
            R2star_map_init,
            R2star_map_init_pre_normalization_vars,
            S0_map_init,
            S0_map_init_pre_normalization_vars,
            B0_map_init,
            B0_map_init_pre_normalization_vars,
            phi_map_init,
            phi_map_init_pre_normalization_vars,
        ) = self.__compute_quantitative_maps__(  # type: ignore
            masked_kspace, sensitivity_map, prediction, anatomy_mask  # type: ignore
        )

        attrs.update(
            self.__parse_normalization_vars__(
                kspace_pre_normalization_vars,  # type: ignore
                sensitivity_pre_normalization_vars,
                prediction_pre_normalization_vars,
                noise_prediction_pre_normalization_vars,
                target_pre_normalization_vars,
                R2star_map_init_pre_normalization_vars,  # type: ignore
                R2star_map_target_pre_normalization_vars,
                S0_map_init_pre_normalization_vars,  # type: ignore
                S0_map_target_pre_normalization_vars,
                B0_map_init_pre_normalization_vars,  # type: ignore
                B0_map_target_pre_normalization_vars,
                phi_map_init_pre_normalization_vars,  # type: ignore
                phi_map_target_pre_normalization_vars,
            )
        )
        attrs["fname"] = fname
        attrs["slice_idx"] = slice_idx

        return (
            R2star_map_init,
            R2star_map_target,
            S0_map_init,
            S0_map_target,
            B0_map_init,
            B0_map_target,
            phi_map_init,
            phi_map_target,
            torch.tensor(self.TEs),
            kspace,
            masked_kspace,  # type: ignore
            sensitivity_map,
            mask,
            anatomy_mask,
            prediction,
            target,
            fname,
            slice_idx,
            acc,  # type: ignore
            attrs,
        )

    def __repr__(self) -> str:
        """Representation of :class:`qMRIDataTransforms`."""
        return (
            f"Preprocessing transforms initialized for {self.__class__.__name__}: "
            f"precompute_quantitative_maps = {self.precompute_quantitative_maps}, "
            f"prewhitening = {self.prewhitening}, "
            f"masking = {self.masking}, "
            f"SSDU masking = {self.ssdu_masking}, "
            f"kspace zero-filling = {self.kspace_zero_filling}, "
            f"cropping = {self.cropping}, "
            f"normalization = {self.normalization}, "
        )

    def __str__(self) -> str:
        """String representation of :class:`qMRIDataTransforms`."""
        return self.__repr__()

    def __process_kspace__(  # noqa: MC0001
        self, kspace: np.ndarray, mask: Union[np.ndarray, None], attrs: Dict, fname: str
    ) -> Tuple[torch.Tensor, Union[List[torch.Tensor], torch.Tensor], Union[List[torch.Tensor], torch.Tensor], int]:
        """Apply the preprocessing transforms to the kspace.

        Parameters
        ----------
        kspace : torch.Tensor
           The kspace.
        mask : torch.Tensor
            The mask, if None, the mask is generated.
        attrs : Dict
            The attributes, if stored in the file.
        fname : str
            The file name.

        Returns
        -------
        Tuple[torch.Tensor, Union[List[torch.Tensor], torch.Tensor], Union[List[torch.Tensor], torch.Tensor], int]
            The transformed (fully-sampled) kspace, the masked kspace, the mask, the attributes and the acceleration
            factor.
        """
        kspace = to_tensor(kspace)
        kspace = add_coil_dim_if_singlecoil(kspace, dim=self.coil_dim)

        kspace_echoes = []
        for ke in range(kspace.shape[0]):
            kspace_echo = kspace[ke]
            kspace_echo = self.coils_shape_transforms(kspace_echo, apply_backward_transform=True)
            kspace_echo = self.prewhitening(kspace_echo)  # type: ignore
            kspace_echoes.append(kspace_echo)
        kspace = torch.stack(kspace_echoes, dim=0)

        if self.crop_before_masking:
            kspace = self.cropping(kspace, apply_backward_transform=not self.kspace_crop)  # type: ignore

        kspace = torch.stack([self.random_motion(kspace[ke]) for ke in range(kspace.shape[0])], dim=0)  # type: ignore

        masked_kspace, mask, acc = self.masking(
            kspace,
            mask,
            (
                attrs["padding_left"] if "padding_left" in attrs else 0,
                attrs["padding_right"] if "padding_right" in attrs else 0,
            ),
            tuple(map(ord, fname)) if self.use_seed else None,  # type: ignore
        )

        if not self.crop_before_masking:
            kspace = self.cropping(kspace, apply_backward_transform=not self.kspace_crop)  # type: ignore
            masked_kspace = self.cropping(masked_kspace, apply_backward_transform=not self.kspace_crop)  # type: ignore
            mask = self.cropping(mask)  # type: ignore

        init_kspace = kspace
        init_masked_kspace = masked_kspace
        init_mask = mask

        if isinstance(kspace, list):
            kspaces = []
            pre_normalization_vars = []
            for i in range(len(kspace)):  # pylint: disable=consider-using-enumerate
                if not is_none(self.normalization.__repr__()):
                    _kspace, _pre_normalization_vars = self.normalization(  # type: ignore
                        kspace[i], apply_backward_transform=True
                    )
                else:
                    _kspace = kspace[i]
                    is_complex = _kspace.shape[-1] == 2
                    if is_complex:
                        _kspace = torch.view_as_complex(_kspace)
                    _pre_normalization_vars = {
                        "min": torch.min(torch.abs(_kspace)),
                        "max": torch.max(torch.abs(_kspace)),
                        "mean": torch.mean(torch.abs(_kspace)),
                        "std": torch.std(torch.abs(_kspace)),
                        "var": torch.var(torch.abs(_kspace)),
                    }
                    if is_complex:
                        _kspace = torch.view_as_real(_kspace)
                kspaces.append(_kspace)
                pre_normalization_vars.append(_pre_normalization_vars)
            kspace = kspaces
        else:
            if not is_none(self.normalization.__repr__()):
                kspace, pre_normalization_vars = self.normalization(  # type: ignore
                    kspace, apply_backward_transform=True
                )
            else:
                is_complex = kspace.shape[-1] == 2
                if is_complex:
                    kspace = torch.view_as_complex(kspace)
                pre_normalization_vars = {  # type: ignore
                    "min": torch.min(torch.abs(kspace)),
                    "max": torch.max(torch.abs(kspace)),
                    "mean": torch.mean(torch.abs(kspace)),
                    "std": torch.std(torch.abs(kspace)),
                    "var": torch.var(torch.abs(kspace)),
                }
                if is_complex:
                    kspace = torch.view_as_real(kspace)

        if isinstance(masked_kspace, list):
            masked_kspaces = []
            masked_pre_normalization_vars = []
            for i in range(len(masked_kspace)):  # pylint: disable=consider-using-enumerate
                if not is_none(self.normalization.__repr__()):
                    _masked_kspace, _masked_pre_normalization_vars = self.normalization(  # type: ignore
                        masked_kspace[i], apply_backward_transform=True
                    )
                else:
                    _masked_kspace = masked_kspace[i]
                    is_complex = _masked_kspace.shape[-1] == 2
                    if is_complex:
                        _masked_kspace = torch.view_as_complex(_masked_kspace)
                    _masked_pre_normalization_vars = {
                        "min": torch.min(torch.abs(_masked_kspace)),
                        "max": torch.max(torch.abs(_masked_kspace)),
                        "mean": torch.mean(torch.abs(_masked_kspace)),
                        "std": torch.std(torch.abs(_masked_kspace)),
                        "var": torch.var(torch.abs(_masked_kspace)),
                    }
                    if is_complex:
                        _masked_kspace = torch.view_as_real(_masked_kspace)
                masked_kspaces.append(_masked_kspace)
                masked_pre_normalization_vars.append(_masked_pre_normalization_vars)
            masked_kspace = masked_kspaces
        else:
            if not is_none(self.normalization.__repr__()):
                masked_kspace, masked_pre_normalization_vars = self.normalization(
                    masked_kspace, apply_backward_transform=True
                )
            else:
                is_complex = masked_kspace.shape[-1] == 2
                if is_complex:
                    masked_kspace = torch.view_as_complex(masked_kspace)
                masked_pre_normalization_vars = {
                    "min": torch.min(torch.abs(masked_kspace)),
                    "max": torch.max(torch.abs(masked_kspace)),
                    "mean": torch.mean(torch.abs(masked_kspace)),
                    "std": torch.std(torch.abs(masked_kspace)),
                    "var": torch.var(torch.abs(masked_kspace)),
                }
                if is_complex:
                    masked_kspace = torch.view_as_real(masked_kspace)

        if self.ssdu:
            kspace, masked_kspace, mask = self.__self_supervised_data_undersampling__(  # type: ignore
                kspace, masked_kspace, mask, fname
            )

        n2r_pre_normalization_vars = None
        if self.n2r and (not attrs["n2r_supervised"] or self.ssdu):
            n2r_masked_kspace, n2r_mask = self.__noise_to_reconstruction__(init_kspace, init_masked_kspace, init_mask)

            if self.ssdu:
                if isinstance(mask, list):
                    for i in range(len(mask)):  # pylint: disable=consider-using-enumerate
                        if init_mask[i].dim() != mask[i][0].dim():  # type: ignore
                            # find dimensions == 1 in mask[i][0] and add them to init_mask
                            unitary_dims = [j for j in range(mask[i][0].dim()) if mask[i][0].shape[j] == 1]
                            # unsqueeze init_mask to the index of the unitary dimensions
                            for j in unitary_dims:
                                init_mask[i] = init_mask[i].unsqueeze(j)  # type: ignore
                        masked_kspace[i] = init_masked_kspace[i]
                        mask[i][0] = init_mask[i]
                else:
                    if init_mask.dim() != mask[0].dim():  # type: ignore
                        # find dimensions == 1 in mask[0] and add them to init_mask
                        unitary_dims = [j for j in range(mask[0].dim()) if mask[0].shape[j] == 1]
                        # unsqueeze init_mask to the index of the unitary dimensions
                        for j in unitary_dims:
                            init_mask = init_mask.unsqueeze(j)  # type: ignore
                    masked_kspace = init_masked_kspace
                    mask[0] = init_mask

            if "None" not in self.normalization.__repr__():
                if isinstance(masked_kspace, list):
                    masked_kspaces = []
                    masked_pre_normalization_vars = []
                    for i in range(len(masked_kspace)):  # pylint: disable=consider-using-enumerate
                        _masked_kspace, _masked_pre_normalization_vars = self.normalization(  # type: ignore
                            masked_kspace[i], apply_backward_transform=True
                        )
                        masked_kspaces.append(_masked_kspace)
                        masked_pre_normalization_vars.append(_masked_pre_normalization_vars)
                    masked_kspace = masked_kspaces
                else:
                    masked_kspace, masked_pre_normalization_vars = self.normalization(  # type: ignore
                        masked_kspace, apply_backward_transform=True
                    )
                if isinstance(n2r_masked_kspace, list):
                    n2r_masked_kspaces = []
                    n2r_pre_normalization_vars = []
                    for i in range(len(n2r_masked_kspace)):  # pylint: disable=consider-using-enumerate
                        _n2r_masked_kspace, _n2r_pre_normalization_vars = self.normalization(  # type: ignore
                            n2r_masked_kspace[i], apply_backward_transform=True
                        )
                        n2r_masked_kspaces.append(_n2r_masked_kspace)
                        n2r_pre_normalization_vars.append(_n2r_pre_normalization_vars)
                    n2r_masked_kspace = n2r_masked_kspaces
                else:
                    n2r_masked_kspace, n2r_pre_normalization_vars = self.normalization(  # type: ignore
                        n2r_masked_kspace, apply_backward_transform=True
                    )
            else:
                masked_pre_normalization_vars = None  # type: ignore
                n2r_pre_normalization_vars = None  # type: ignore

            masked_kspace = [masked_kspace, n2r_masked_kspace]
            mask = [mask, n2r_mask]

        if self.normalization_type == "grayscale":
            if isinstance(mask, list):
                masks = []
                for i in range(len(mask)):  # pylint: disable=consider-using-enumerate
                    _mask, _ = self.normalization(mask[i], apply_backward_transform=False)  # type: ignore
                    masks.append(_mask)
                mask = masks
            else:
                mask, _ = self.normalization(mask, apply_backward_transform=False)  # type: ignore

        pre_normalization_vars = {  # type: ignore
            "kspace_pre_normalization_vars": pre_normalization_vars,
            "masked_kspace_pre_normalization_vars": masked_pre_normalization_vars,
            "noise_masked_kspace_pre_normalization_vars": n2r_pre_normalization_vars,
        }

        return kspace, masked_kspace, mask, pre_normalization_vars, acc  # type: ignore

    def __noise_to_reconstruction__(
        self,
        kspace: torch.Tensor,
        masked_kspace: torch.Tensor,
        mask: Union[List, torch.Tensor],
    ) -> Tuple[Union[List, torch.Tensor], Union[List, torch.Tensor]]:
        """Apply the noise-to-reconstruction transform.

        Parameters
        ----------
        kspace : torch.Tensor
            The fully-sampled kspace.
        masked_kspace : torch.Tensor
            The undersampled kspace.
        mask : Union[List, torch.Tensor]
            The undersampling mask.

        Returns
        -------
        n2r_masked_kspace : Union[List, torch.Tensor]
            The noise-to-reconstruction undersampled kspace.
        n2r_mask : Union[List, torch.Tensor]
            The noise-to-reconstruction mask.
        """
        if isinstance(mask, list):
            n2r_masked_kspaces = []
            n2r_masks = []
            for i in range(len(mask)):  # pylint: disable=consider-using-enumerate
                n2r_mask = self.n2r_masking(kspace, mask[i])  # type: ignore  # pylint: disable=not-callable
                n2r_masks.append(n2r_mask)
                n2r_masked_kspaces.append(masked_kspace[i] * n2r_mask + 0.0)
            n2r_mask = n2r_masks
            n2r_masked_kspace = n2r_masked_kspaces
        else:
            n2r_mask = self.n2r_masking(kspace, mask)  # type: ignore  # pylint: disable=not-callable
            n2r_masked_kspace = masked_kspace * n2r_mask + 0.0
        return n2r_masked_kspace, n2r_mask

    def __self_supervised_data_undersampling__(  # noqa: MC0001
        self,
        kspace: torch.Tensor,
        masked_kspace: Union[List, torch.Tensor],
        mask: Union[List, torch.Tensor],
        fname: str,
    ) -> Tuple[
        List[float | Any] | float | Any,
        List[float | Any] | float | Any,
        List[List[torch.Tensor | Any]] | List[torch.Tensor | Any],
    ]:
        """Self-supervised data undersampling.

        Parameters
        ----------
        kspace : torch.Tensor
            The fully-sampled kspace.
        masked_kspace : Union[List, torch.Tensor]
            The undersampled kspace.
        mask : Union[List, torch.Tensor]
            The undersampling mask.
        fname : str
            The filename of the current sample.

        Returns
        -------
        kspace : torch.Tensor
            The kspace with the loss mask applied.
        masked_kspace : torch.Tensor
            The kspace with the train mask applied.
        mask : list, [torch.Tensor, torch.Tensor]
            The train and loss masks.
        """
        if isinstance(mask, list):
            kspaces = []
            masked_kspaces = []
            masks = []
            for i in range(len(mask)):  # pylint: disable=consider-using-enumerate
                is_1d = mask[i].squeeze().dim() == 1
                if self.shift_mask:
                    mask[i] = torch.fft.fftshift(mask[i].squeeze(-1), dim=(-2, -1)).unsqueeze(-1)
                mask[i] = mask[i].squeeze()
                if is_1d:
                    mask[i] = mask[i].unsqueeze(0).repeat_interleave(kspace.shape[1], dim=0)
                train_mask, loss_mask = self.ssdu_masking(  # type: ignore  # pylint: disable=not-callable
                    kspace, mask[i], fname
                )
                if self.shift_mask:
                    train_mask = torch.fft.fftshift(train_mask, dim=(0, 1))
                    loss_mask = torch.fft.fftshift(loss_mask, dim=(0, 1))
                if is_1d:
                    train_mask = train_mask.unsqueeze(0).unsqueeze(-1)
                    loss_mask = loss_mask.unsqueeze(0).unsqueeze(-1)
                else:
                    # find unitary dims in mask
                    dims = [i for i, x in enumerate(mask[i].shape) if x == 1]
                    # unsqueeze to broadcast
                    for d in dims:
                        train_mask = train_mask.unsqueeze(d)
                        loss_mask = loss_mask.unsqueeze(d)
                if train_mask.dim() != kspace.dim():
                    # find dims != to any train_mask dim
                    dims = [i for i, x in enumerate(kspace.shape) if x not in train_mask.shape]
                    # unsqueeze to broadcast
                    for d in dims:
                        train_mask = train_mask.unsqueeze(d)
                        loss_mask = loss_mask.unsqueeze(d)
                kspaces.append(kspace * loss_mask + 0.0)
                masked_kspaces.append(masked_kspace[i] * train_mask + 0.0)
                masks.append([train_mask, loss_mask])
            kspace = kspaces
            masked_kspace = masked_kspaces
            mask = masks
        else:
            is_1d = mask.squeeze().dim() == 1
            if self.shift_mask:
                mask = torch.fft.fftshift(mask.squeeze(-1), dim=(-2, -1)).unsqueeze(-1)
            mask = mask.squeeze()
            if is_1d:
                mask = mask.unsqueeze(0).repeat_interleave(kspace.shape[1], dim=0)
            train_mask, loss_mask = self.ssdu_masking(  # type: ignore  # pylint: disable=not-callable
                kspace, mask, fname
            )
            if self.shift_mask:
                train_mask = torch.fft.fftshift(train_mask, dim=(0, 1))
                loss_mask = torch.fft.fftshift(loss_mask, dim=(0, 1))
            if is_1d:
                train_mask = train_mask.unsqueeze(0).unsqueeze(-1)
                loss_mask = loss_mask.unsqueeze(0).unsqueeze(-1)
            else:
                # find unitary dims in mask
                dims = [i for i, x in enumerate(mask.shape) if x == 1]
                # unsqueeze to broadcast
                for d in dims:
                    train_mask = train_mask.unsqueeze(d)
                    loss_mask = loss_mask.unsqueeze(d)
            if train_mask.dim() != kspace.dim():
                # find dims != to any train_mask dim
                dims = [i for i, x in enumerate(kspace.shape) if x not in train_mask.shape]
                # unsqueeze to broadcast
                for d in dims:
                    train_mask = train_mask.unsqueeze(d)
                    loss_mask = loss_mask.unsqueeze(d)
            kspace = kspace * loss_mask + 0.0
            masked_kspace = masked_kspace * train_mask + 0.0
            mask = [train_mask, loss_mask]
        return kspace, masked_kspace, mask

    def __process_coil_sensitivities_map__(self, sensitivity_map: np.ndarray, kspace: torch.Tensor) -> torch.Tensor:
        """Preprocesses the coil sensitivities map.

        Parameters
        ----------
        sensitivity_map : np.ndarray
            The coil sensitivities map.
        kspace : torch.Tensor
            The kspace.

        Returns
        -------
        torch.Tensor
            The preprocessed coil sensitivities map.
        """
        # This condition is necessary in case of auto estimation of sense maps.
        if self.coil_sensitivity_maps_estimator is not None:
            sensitivity_map = self.coil_sensitivity_maps_estimator(kspace)
        elif sensitivity_map is not None and sensitivity_map.size != 0:
            sensitivity_map = to_tensor(sensitivity_map)
            sensitivity_map = self.coils_shape_transforms(sensitivity_map, apply_forward_transform=True)
            sensitivity_map = self.cropping(sensitivity_map, apply_forward_transform=self.kspace_crop)  # type: ignore
        else:
            # If no sensitivity map is provided, either the data is singlecoil or the sense net is used.
            # Initialize the sensitivity map to 1 to assure for the singlecoil case.
            sensitivity_map = torch.ones_like(kspace) if not isinstance(kspace, list) else torch.ones_like(kspace[0])
        if not is_none(self.normalization.__repr__()):
            sensitivity_map, pre_normalization_vars = self.normalization(  # type: ignore
                sensitivity_map, apply_forward_transform=self.kspace_crop
            )
        else:
            is_complex = sensitivity_map.shape[-1] == 2
            if is_complex:
                sensitivity_map = torch.view_as_complex(sensitivity_map)
            pre_normalization_vars = {
                "min": torch.min(torch.abs(sensitivity_map)),
                "max": torch.max(torch.abs(sensitivity_map)),
                "mean": torch.mean(torch.abs(sensitivity_map)),
                "std": torch.std(torch.abs(sensitivity_map)),
                "var": torch.var(torch.abs(sensitivity_map)),
            }
            if is_complex:
                sensitivity_map = torch.view_as_real(sensitivity_map)
        return sensitivity_map, pre_normalization_vars

    def __compute_quantitative_maps__(
        self,
        kspace: torch.Tensor,
        sensitivity_map: torch.Tensor,
        prediction: Union[torch.Tensor, None],
        anatomy_mask: torch.Tensor,
    ) -> Tuple[
        List[Any] | Any,
        List[Dict[str, torch.Tensor]],
        List[Any] | Any,
        List[Dict[str, torch.Tensor]],
        List[Any] | Any,
        List[Dict[str, torch.Tensor]],
        List[Any] | Any,
        List[Dict[str, torch.Tensor]],
    ]:
        """Compute quantitative maps from the masked kspace data.

        Parameters
        ----------
        kspace: torch.Tensor
            Kspace data.
        sensitivity_map: torch.Tensor
            Sensitivity maps.
        prediction: Union[torch.Tensor, None]
            Initial prediction or None.
        anatomy_mask: torch.Tensor
            Brain mask.

        Returns
        -------
        R2star_map: torch.Tensor
            Computed R2* map.
        R2star_map_pre_normalization_vars: Dict[str, torch.Tensor]
            Computed R2* map pre-normalization variables.
        S0_map: torch.Tensor
            Computed S0 map.
        S0_map_pre_normalization_vars: Dict[str, torch.Tensor]
            Computed S0 map pre-normalization variables.
        B0_map: torch.Tensor
            Computed B0 map.
        B0_map_pre_normalization_vars: Dict[str, torch.Tensor]
            Computed B0 map pre-normalization variables.
        phi_map: torch.Tensor
            Computed phi map.
        phi_map_pre_normalization_vars: Dict[str, torch.Tensor]
            Computed phi map pre-normalization variables.
        """
        R2star_maps = []
        R2star_map_pre_normalization_vars = []
        S0_maps = []
        S0_map_pre_normalization_vars = []
        B0_maps = []
        B0_map_pre_normalization_vars = []
        phi_maps = []
        phi_map_pre_normalization_vars = []
        if isinstance(kspace, list):
            for i, _ in enumerate(kspace):
                R2star_map, S0_map, B0_map, phi_map = R2star_B0_S0_phi_mapping(
                    prediction[i] if isinstance(prediction, list) else prediction,
                    self.TEs,
                    anatomy_mask,
                    scaling_factor=self.qmaps_scaling_factor,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )

                R2star_maps.append(R2star_map)
                R2star_map_pre_normalization_vars.append(
                    {
                        "min": torch.min(torch.abs(R2star_map)),
                        "max": torch.max(torch.abs(R2star_map)),
                        "mean": torch.mean(torch.abs(R2star_map)),
                        "std": torch.std(torch.abs(R2star_map)),
                        "var": torch.var(torch.abs(R2star_map)),
                    }
                )
                S0_maps.append(S0_map)
                S0_map_pre_normalization_vars.append(
                    {
                        "min": torch.min(torch.abs(S0_map)),
                        "max": torch.max(torch.abs(S0_map)),
                        "mean": torch.mean(torch.abs(S0_map)),
                        "std": torch.std(torch.abs(S0_map)),
                        "var": torch.var(torch.abs(S0_map)),
                    }
                )
                B0_maps.append(B0_map)
                B0_map_pre_normalization_vars.append(
                    {
                        "min": torch.min(torch.abs(B0_map)),
                        "max": torch.max(torch.abs(B0_map)),
                        "mean": torch.mean(torch.abs(B0_map)),
                        "std": torch.std(torch.abs(B0_map)),
                        "var": torch.var(torch.abs(B0_map)),
                    }
                )
                phi_maps.append(phi_map)
                phi_map_pre_normalization_vars.append(
                    {
                        "min": torch.min(torch.abs(phi_map)),
                        "max": torch.max(torch.abs(phi_map)),
                        "mean": torch.mean(torch.abs(phi_map)),
                        "std": torch.std(torch.abs(phi_map)),
                        "var": torch.var(torch.abs(phi_map)),
                    }
                )

            R2star_map = R2star_maps
            S0_map = S0_maps
            B0_map = B0_maps
            phi_map = phi_maps
        else:
            if prediction is None:
                prediction = sense(
                    ifft2(
                        kspace,
                        centered=self.fft_centered,
                        normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                    ),
                    sensitivity_map.unsqueeze(0),
                    dim=self.coil_dim,
                )

            R2star_map, S0_map, B0_map, phi_map = R2star_B0_S0_phi_mapping(
                prediction,
                self.TEs,
                anatomy_mask,
                scaling_factor=self.qmaps_scaling_factor,
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )

            R2star_map_pre_normalization_vars = {  # type: ignore
                "min": torch.min(torch.abs(R2star_map)),
                "max": torch.max(torch.abs(R2star_map)),
                "mean": torch.mean(torch.abs(R2star_map)),
                "std": torch.std(torch.abs(R2star_map)),
                "var": torch.var(torch.abs(R2star_map)),
            }
            S0_map_pre_normalization_vars = {  # type: ignore
                "min": torch.min(torch.abs(S0_map)),
                "max": torch.max(torch.abs(S0_map)),
                "mean": torch.mean(torch.abs(S0_map)),
                "std": torch.std(torch.abs(S0_map)),
                "var": torch.var(torch.abs(S0_map)),
            }
            B0_map_pre_normalization_vars = {  # type: ignore
                "min": torch.min(torch.abs(B0_map)),
                "max": torch.max(torch.abs(B0_map)),
                "mean": torch.mean(torch.abs(B0_map)),
                "std": torch.std(torch.abs(B0_map)),
                "var": torch.var(torch.abs(B0_map)),
            }
            phi_map_pre_normalization_vars = {  # type: ignore
                "min": torch.min(torch.abs(phi_map)),
                "max": torch.max(torch.abs(phi_map)),
                "mean": torch.mean(torch.abs(phi_map)),
                "std": torch.std(torch.abs(phi_map)),
                "var": torch.var(torch.abs(phi_map)),
            }

        return (
            R2star_map,
            R2star_map_pre_normalization_vars,
            S0_map,
            S0_map_pre_normalization_vars,
            B0_map,
            B0_map_pre_normalization_vars,
            phi_map,
            phi_map_pre_normalization_vars,
        )

    def __initialize_prediction__(
        self, prediction: Union[torch.Tensor, np.ndarray, None], kspace: torch.Tensor, sensitivity_map: torch.Tensor
    ) -> Union[List[torch.Tensor], torch.Tensor]:
        """Predicts a coil-combined image.

        Parameters
        ----------
        prediction : np.ndarray
            The initial estimation, if None, the prediction is initialized.
        kspace : torch.Tensor
            The kspace.
        sensitivity_map : torch.Tensor
            The sensitivity map.

        Returns
        -------
        Union[List[torch.Tensor], torch.Tensor]
            The initialized prediction, either a list of coil-combined images or a single coil-combined image.
        """
        if is_none(prediction) or prediction.ndim < 2:  # type: ignore
            if isinstance(kspace, list):
                prediction = []
                pre_normalization_vars = []
                for y in kspace:
                    pred = coil_combination_method_func(
                        ifft2(y, self.fft_centered, self.fft_normalization, self.spatial_dims),
                        sensitivity_map,
                        method=self.coil_combination_method,
                        dim=self.coil_dim,
                    )
                    pred = self.cropping(pred, apply_forward_transform=self.kspace_crop)  # type: ignore
                    if not is_none(self.normalization.__repr__()):
                        pred, _pre_normalization_vars = self.normalization(  # type: ignore
                            pred, apply_forward_transform=self.kspace_crop
                        )
                    else:
                        if pred.shape[-1] == 2:
                            pred = torch.view_as_complex(pred)
                        _pre_normalization_vars = {
                            "min": torch.min(torch.abs(pred)),
                            "max": torch.max(torch.abs(pred)),
                            "mean": torch.mean(torch.abs(pred)),
                            "std": torch.std(torch.abs(pred)),
                            "var": torch.var(torch.abs(pred)),
                        }
                    prediction.append(pred)
                    pre_normalization_vars.append(_pre_normalization_vars)
                if prediction[0].shape[-1] != 2 and torch.is_complex(prediction[0]):
                    prediction = [torch.view_as_real(x) for x in prediction]
            else:
                prediction = coil_combination_method_func(
                    ifft2(kspace, self.fft_centered, self.fft_normalization, self.spatial_dims),
                    sensitivity_map,
                    method=self.coil_combination_method,
                    dim=self.coil_dim,
                )
                prediction = self.cropping(prediction, apply_forward_transform=self.kspace_crop)  # type: ignore
                if not is_none(self.normalization.__repr__()):
                    prediction, pre_normalization_vars = self.normalization(  # type: ignore
                        prediction, apply_forward_transform=self.kspace_crop
                    )
                else:
                    if prediction.shape[-1] == 2:
                        prediction = torch.view_as_complex(prediction)
                    pre_normalization_vars = {  # type: ignore
                        "min": torch.min(torch.abs(prediction)),
                        "max": torch.max(torch.abs(prediction)),
                        "mean": torch.mean(torch.abs(prediction)),
                        "std": torch.std(torch.abs(prediction)),
                        "var": torch.var(torch.abs(prediction)),
                    }
                if prediction.shape[-1] != 2 and torch.is_complex(prediction):
                    prediction = torch.view_as_real(prediction)
        else:
            if isinstance(prediction, np.ndarray):
                prediction = to_tensor(prediction)
            prediction = self.cropping(prediction, apply_forward_transform=self.kspace_crop)  # type: ignore
            if not is_none(self.normalization.__repr__()):
                prediction, pre_normalization_vars = self.normalization(  # type: ignore
                    prediction, apply_forward_transform=self.kspace_crop
                )
            else:
                if prediction.shape[-1] == 2:  # type: ignore
                    prediction = torch.view_as_complex(prediction)
                pre_normalization_vars = {  # type: ignore
                    "min": torch.min(torch.abs(prediction)),
                    "max": torch.max(torch.abs(prediction)),
                    "mean": torch.mean(torch.abs(prediction)),
                    "std": torch.std(torch.abs(prediction)),
                    "var": torch.var(torch.abs(prediction)),
                }
            if prediction.shape[-1] != 2 and torch.is_complex(prediction):
                prediction = torch.view_as_real(prediction)

        return prediction, pre_normalization_vars

    @staticmethod
    def __parse_normalization_vars__(  # noqa: MC0001
        kspace_vars,
        sensitivity_vars,
        prediction_vars,
        noise_prediction_vars,
        target_vars,
        R2star_maps_init_vars,
        R2star_map_target_vars,
        S0_map_init_vars,
        S0_map_target_vars,
        B0_map_init_vars,
        B0_map_target_vars,
        phi_map_init_vars,
        phi_map_target_vars,
    ) -> Dict:
        """Parses the normalization variables and returns a unified dictionary.

        Parameters
        ----------
        kspace_vars : Dict
            The kspace normalization variables.
        sensitivity_vars : Dict
            The sensitivity map normalization variables.
        prediction_vars : Dict
            The prediction normalization variables.
        noise_prediction_vars : Union[Dict, None]
            The noise prediction normalization variables.
        target_vars : Dict
            The target normalization variables.
        R2star_maps_init_vars : Dict
            The R2* maps initialization normalization variables.
        R2star_map_target_vars : Dict
            The R2* maps target normalization variables.
        S0_map_init_vars : Dict
            The S0 maps initialization normalization variables.
        S0_map_target_vars : Dict
            The S0 maps target normalization variables.
        B0_map_init_vars : Dict
            The B0 maps initialization normalization variables.
        B0_map_target_vars : Dict
            The B0 maps target normalization variables.
        phi_map_init_vars : Dict
            The phi maps initialization normalization variables.
        phi_map_target_vars : Dict
            The phi maps target normalization variables.

        Returns
        -------
        Dict
            The normalization variables.
        """
        normalization_vars = {}

        masked_kspace_vars = kspace_vars["masked_kspace_pre_normalization_vars"]
        if isinstance(masked_kspace_vars, list):
            if masked_kspace_vars[0] is not None:
                for i, masked_kspace_var in enumerate(masked_kspace_vars):
                    normalization_vars[f"masked_kspace_min_{i}"] = masked_kspace_var["min"]
                    normalization_vars[f"masked_kspace_max_{i}"] = masked_kspace_var["max"]
                    normalization_vars[f"masked_kspace_mean_{i}"] = masked_kspace_var["mean"]
                    normalization_vars[f"masked_kspace_std_{i}"] = masked_kspace_var["std"]
                    normalization_vars[f"masked_kspace_var_{i}"] = masked_kspace_var["var"]
        else:
            if masked_kspace_vars is not None:
                normalization_vars["masked_kspace_min"] = masked_kspace_vars["min"]
                normalization_vars["masked_kspace_max"] = masked_kspace_vars["max"]
                normalization_vars["masked_kspace_mean"] = masked_kspace_vars["mean"]
                normalization_vars["masked_kspace_std"] = masked_kspace_vars["std"]
                normalization_vars["masked_kspace_var"] = masked_kspace_vars["var"]

        noise_masked_kspace_vars = kspace_vars["noise_masked_kspace_pre_normalization_vars"]
        if noise_masked_kspace_vars is not None:
            if isinstance(noise_masked_kspace_vars, list):
                if noise_masked_kspace_vars[0] is not None:
                    for i, noise_masked_kspace_var in enumerate(noise_masked_kspace_vars):
                        normalization_vars[f"noise_masked_kspace_min_{i}"] = noise_masked_kspace_var["min"]
                        normalization_vars[f"noise_masked_kspace_max_{i}"] = noise_masked_kspace_var["max"]
                        normalization_vars[f"noise_masked_kspace_mean_{i}"] = noise_masked_kspace_var["mean"]
                        normalization_vars[f"noise_masked_kspace_std_{i}"] = noise_masked_kspace_var["std"]
                        normalization_vars[f"noise_masked_kspace_var_{i}"] = noise_masked_kspace_var["var"]
            else:
                if noise_masked_kspace_vars is not None:
                    normalization_vars["noise_masked_kspace_min"] = noise_masked_kspace_vars["min"]
                    normalization_vars["noise_masked_kspace_max"] = noise_masked_kspace_vars["max"]
                    normalization_vars["noise_masked_kspace_mean"] = noise_masked_kspace_vars["mean"]
                    normalization_vars["noise_masked_kspace_std"] = noise_masked_kspace_vars["std"]
                    normalization_vars["noise_masked_kspace_var"] = noise_masked_kspace_vars["var"]

        kspace_vars = kspace_vars["kspace_pre_normalization_vars"]
        if isinstance(kspace_vars, list):
            if kspace_vars[0] is not None:
                for i, kspace_var in enumerate(kspace_vars):
                    normalization_vars[f"kspace_min_{i}"] = kspace_var["min"]
                    normalization_vars[f"kspace_max_{i}"] = kspace_var["max"]
                    normalization_vars[f"kspace_mean_{i}"] = kspace_var["mean"]
                    normalization_vars[f"kspace_std_{i}"] = kspace_var["std"]
                    normalization_vars[f"kspace_var_{i}"] = kspace_var["var"]
        else:
            if kspace_vars is not None:
                normalization_vars["kspace_min"] = kspace_vars["min"]
                normalization_vars["kspace_max"] = kspace_vars["max"]
                normalization_vars["kspace_mean"] = kspace_vars["mean"]
                normalization_vars["kspace_std"] = kspace_vars["std"]
                normalization_vars["kspace_var"] = kspace_vars["var"]

        if sensitivity_vars is not None:
            normalization_vars["sensitivity_maps_min"] = sensitivity_vars["min"]
            normalization_vars["sensitivity_maps_max"] = sensitivity_vars["max"]
            normalization_vars["sensitivity_maps_mean"] = sensitivity_vars["mean"]
            normalization_vars["sensitivity_maps_std"] = sensitivity_vars["std"]
            normalization_vars["sensitivity_maps_var"] = sensitivity_vars["var"]

        if isinstance(prediction_vars, list):
            if prediction_vars[0] is not None:
                for i, prediction_var in enumerate(prediction_vars):
                    normalization_vars[f"prediction_min_{i}"] = prediction_var["min"]
                    normalization_vars[f"prediction_max_{i}"] = prediction_var["max"]
                    normalization_vars[f"prediction_mean_{i}"] = prediction_var["mean"]
                    normalization_vars[f"prediction_std_{i}"] = prediction_var["std"]
                    normalization_vars[f"prediction_var_{i}"] = prediction_var["var"]
        else:
            if prediction_vars is not None:
                normalization_vars["prediction_min"] = prediction_vars["min"]
                normalization_vars["prediction_max"] = prediction_vars["max"]
                normalization_vars["prediction_mean"] = prediction_vars["mean"]
                normalization_vars["prediction_std"] = prediction_vars["std"]
                normalization_vars["prediction_var"] = prediction_vars["var"]

        if noise_prediction_vars is not None:
            if isinstance(noise_prediction_vars, list):
                for i, noise_prediction_var in enumerate(noise_prediction_vars):
                    normalization_vars[f"noise_prediction_min_{i}"] = noise_prediction_var["min"]
                    normalization_vars[f"noise_prediction_max_{i}"] = noise_prediction_var["max"]
                    normalization_vars[f"noise_prediction_mean_{i}"] = noise_prediction_var["mean"]
                    normalization_vars[f"noise_prediction_std_{i}"] = noise_prediction_var["std"]
                    normalization_vars[f"noise_prediction_var_{i}"] = noise_prediction_var["var"]
            else:
                normalization_vars["noise_prediction_min"] = noise_prediction_vars["min"]
                normalization_vars["noise_prediction_max"] = noise_prediction_vars["max"]
                normalization_vars["noise_prediction_mean"] = noise_prediction_vars["mean"]
                normalization_vars["noise_prediction_std"] = noise_prediction_vars["std"]
                normalization_vars["noise_prediction_var"] = noise_prediction_vars["var"]

        if isinstance(target_vars, list):
            if target_vars[0] is not None:
                for i, target_var in enumerate(target_vars):
                    normalization_vars[f"target_min_{i}"] = target_var["min"]
                    normalization_vars[f"target_max_{i}"] = target_var["max"]
                    normalization_vars[f"target_mean_{i}"] = target_var["mean"]
                    normalization_vars[f"target_std_{i}"] = target_var["std"]
                    normalization_vars[f"target_var_{i}"] = target_var["var"]
        else:
            if target_vars is not None:
                normalization_vars["target_min"] = target_vars["min"]
                normalization_vars["target_max"] = target_vars["max"]
                normalization_vars["target_mean"] = target_vars["mean"]
                normalization_vars["target_std"] = target_vars["std"]
                normalization_vars["target_var"] = target_vars["var"]

        if isinstance(R2star_maps_init_vars, list):
            if R2star_maps_init_vars[0] is not None:
                for i, R2star_map_init_var in enumerate(R2star_maps_init_vars):
                    normalization_vars[f"R2star_map_init_min_{i}"] = R2star_map_init_var["min"]
                    normalization_vars[f"R2star_map_init_max_{i}"] = R2star_map_init_var["max"]
                    normalization_vars[f"R2star_map_init_mean_{i}"] = R2star_map_init_var["mean"]
                    normalization_vars[f"R2star_map_init_std_{i}"] = R2star_map_init_var["std"]
                    normalization_vars[f"R2star_map_init_var_{i}"] = R2star_map_init_var["var"]
        else:
            if R2star_maps_init_vars is not None:
                normalization_vars["R2star_map_init_min"] = R2star_maps_init_vars["min"]
                normalization_vars["R2star_map_init_max"] = R2star_maps_init_vars["max"]
                normalization_vars["R2star_map_init_mean"] = R2star_maps_init_vars["mean"]
                normalization_vars["R2star_map_init_std"] = R2star_maps_init_vars["std"]
                normalization_vars["R2star_map_init_var"] = R2star_maps_init_vars["var"]

        if isinstance(R2star_map_target_vars, list):
            if R2star_map_target_vars[0] is not None:
                for i, R2star_map_target_var in enumerate(R2star_map_target_vars):
                    normalization_vars[f"R2star_map_target_min_{i}"] = R2star_map_target_var["min"]
                    normalization_vars[f"R2star_map_target_max_{i}"] = R2star_map_target_var["max"]
                    normalization_vars[f"R2star_map_target_mean_{i}"] = R2star_map_target_var["mean"]
                    normalization_vars[f"R2star_map_target_std_{i}"] = R2star_map_target_var["std"]
                    normalization_vars[f"R2star_map_target_var_{i}"] = R2star_map_target_var["var"]
        else:
            if R2star_map_target_vars is not None:
                normalization_vars["R2star_map_target_min"] = R2star_map_target_vars["min"]
                normalization_vars["R2star_map_target_max"] = R2star_map_target_vars["max"]
                normalization_vars["R2star_map_target_mean"] = R2star_map_target_vars["mean"]
                normalization_vars["R2star_map_target_std"] = R2star_map_target_vars["std"]
                normalization_vars["R2star_map_target_var"] = R2star_map_target_vars["var"]

        if isinstance(S0_map_init_vars, list):
            if S0_map_init_vars[0] is not None:
                for i, S0_map_init_var in enumerate(S0_map_init_vars):
                    normalization_vars[f"S0_map_init_min_{i}"] = S0_map_init_var["min"]
                    normalization_vars[f"S0_map_init_max_{i}"] = S0_map_init_var["max"]
                    normalization_vars[f"S0_map_init_mean_{i}"] = S0_map_init_var["mean"]
                    normalization_vars[f"S0_map_init_std_{i}"] = S0_map_init_var["std"]
                    normalization_vars[f"S0_map_init_var_{i}"] = S0_map_init_var["var"]
        else:
            if S0_map_init_vars is not None:
                normalization_vars["S0_map_init_min"] = S0_map_init_vars["min"]
                normalization_vars["S0_map_init_max"] = S0_map_init_vars["max"]
                normalization_vars["S0_map_init_mean"] = S0_map_init_vars["mean"]
                normalization_vars["S0_map_init_std"] = S0_map_init_vars["std"]
                normalization_vars["S0_map_init_var"] = S0_map_init_vars["var"]

        if isinstance(S0_map_target_vars, list):
            if S0_map_target_vars[0] is not None:
                for i, S0_map_target_var in enumerate(S0_map_target_vars):
                    normalization_vars[f"S0_map_target_min_{i}"] = S0_map_target_var["min"]
                    normalization_vars[f"S0_map_target_max_{i}"] = S0_map_target_var["max"]
                    normalization_vars[f"S0_map_target_mean_{i}"] = S0_map_target_var["mean"]
                    normalization_vars[f"S0_map_target_std_{i}"] = S0_map_target_var["std"]
                    normalization_vars[f"S0_map_target_var_{i}"] = S0_map_target_var["var"]
        else:
            if S0_map_target_vars is not None:
                normalization_vars["S0_map_target_min"] = S0_map_target_vars["min"]
                normalization_vars["S0_map_target_max"] = S0_map_target_vars["max"]
                normalization_vars["S0_map_target_mean"] = S0_map_target_vars["mean"]
                normalization_vars["S0_map_target_std"] = S0_map_target_vars["std"]
                normalization_vars["S0_map_target_var"] = S0_map_target_vars["var"]

        if isinstance(B0_map_init_vars, list):
            if B0_map_init_vars[0] is not None:
                for i, B0_map_init_var in enumerate(B0_map_init_vars):
                    normalization_vars[f"B0_map_init_min_{i}"] = B0_map_init_var["min"]
                    normalization_vars[f"B0_map_init_max_{i}"] = B0_map_init_var["max"]
                    normalization_vars[f"B0_map_init_mean_{i}"] = B0_map_init_var["mean"]
                    normalization_vars[f"B0_map_init_std_{i}"] = B0_map_init_var["std"]
                    normalization_vars[f"B0_map_init_var_{i}"] = B0_map_init_var["var"]
        else:
            if B0_map_init_vars is not None:
                normalization_vars["B0_map_init_min"] = B0_map_init_vars["min"]
                normalization_vars["B0_map_init_max"] = B0_map_init_vars["max"]
                normalization_vars["B0_map_init_mean"] = B0_map_init_vars["mean"]
                normalization_vars["B0_map_init_std"] = B0_map_init_vars["std"]
                normalization_vars["B0_map_init_var"] = B0_map_init_vars["var"]

        if isinstance(B0_map_target_vars, list):
            if B0_map_target_vars[0] is not None:
                for i, B0_map_target_var in enumerate(B0_map_target_vars):
                    normalization_vars[f"B0_map_target_min_{i}"] = B0_map_target_var["min"]
                    normalization_vars[f"B0_map_target_max_{i}"] = B0_map_target_var["max"]
                    normalization_vars[f"B0_map_target_mean_{i}"] = B0_map_target_var["mean"]
                    normalization_vars[f"B0_map_target_std_{i}"] = B0_map_target_var["std"]
                    normalization_vars[f"B0_map_target_var_{i}"] = B0_map_target_var["var"]
        else:
            if B0_map_target_vars is not None:
                normalization_vars["B0_map_target_min"] = B0_map_target_vars["min"]
                normalization_vars["B0_map_target_max"] = B0_map_target_vars["max"]
                normalization_vars["B0_map_target_mean"] = B0_map_target_vars["mean"]
                normalization_vars["B0_map_target_std"] = B0_map_target_vars["std"]
                normalization_vars["B0_map_target_var"] = B0_map_target_vars["var"]

        if isinstance(phi_map_init_vars, list):
            if phi_map_init_vars[0] is not None:
                for i, phi_map_init_var in enumerate(phi_map_init_vars):
                    normalization_vars[f"phi_map_init_min_{i}"] = phi_map_init_var["min"]
                    normalization_vars[f"phi_map_init_max_{i}"] = phi_map_init_var["max"]
                    normalization_vars[f"phi_map_init_mean_{i}"] = phi_map_init_var["mean"]
                    normalization_vars[f"phi_map_init_std_{i}"] = phi_map_init_var["std"]
                    normalization_vars[f"phi_map_init_var_{i}"] = phi_map_init_var["var"]
        else:
            if phi_map_init_vars is not None:
                normalization_vars["phi_map_init_min"] = phi_map_init_vars["min"]
                normalization_vars["phi_map_init_max"] = phi_map_init_vars["max"]
                normalization_vars["phi_map_init_mean"] = phi_map_init_vars["mean"]
                normalization_vars["phi_map_init_std"] = phi_map_init_vars["std"]
                normalization_vars["phi_map_init_var"] = phi_map_init_vars["var"]

        if isinstance(phi_map_target_vars, list):
            if phi_map_target_vars[0] is not None:
                for i, phi_map_target_var in enumerate(phi_map_target_vars):
                    normalization_vars[f"phi_map_target_min_{i}"] = phi_map_target_var["min"]
                    normalization_vars[f"phi_map_target_max_{i}"] = phi_map_target_var["max"]
                    normalization_vars[f"phi_map_target_mean_{i}"] = phi_map_target_var["mean"]
                    normalization_vars[f"phi_map_target_std_{i}"] = phi_map_target_var["std"]
                    normalization_vars[f"phi_map_target_var_{i}"] = phi_map_target_var["var"]
        else:
            if phi_map_target_vars is not None:
                normalization_vars["phi_map_target_min"] = phi_map_target_vars["min"]
                normalization_vars["phi_map_target_max"] = phi_map_target_vars["max"]
                normalization_vars["phi_map_target_mean"] = phi_map_target_vars["mean"]
                normalization_vars["phi_map_target_std"] = phi_map_target_vars["std"]
                normalization_vars["phi_map_target_var"] = phi_map_target_vars["var"]

        return normalization_vars


class GaussianSmoothing(torch.nn.Module):
    """Apply gaussian smoothing on a 1d, 2d or 3d tensor. Filtering is performed separately for each channel in the
    input using a depthwise convolution.
    """

    def __init__(
        self,
        channels: int,
        kernel_size: Union[List[int], int],
        sigma: float,
        dim: int = 2,
        shift: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
    ):
        """Inits :class:`GaussianSmoothing`.

        Parameters
        ----------
        channels : int
            Number of channels in the input tensor.
        kernel_size : Union[Optional[List[int]], int]
            Gaussian kernel size.
        sigma : float
            Gaussian kernel standard deviation.
        dim : int
            Number of dimensions in the input tensor.
        shift : bool
            If True, the gaussian kernel is centered at (kernel_size - 1) / 2.
        fft_centered : bool
            Whether to center the FFT for a real- or complex-valued input.
        fft_normalization : str
            Whether to normalize the FFT output (None, "ortho", "backward", "forward", "none").
        spatial_dims : Sequence[int]
            Spatial dimensions to keep in the FFT.
        """
        super().__init__()

        self.shift = shift
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims

        if isinstance(kernel_size, int):
            kernel_size = [kernel_size] * dim

        if isinstance(sigma, float):
            sigma = [sigma] * dim  # type: ignore

        # The gaussian kernel is the product of the gaussian function of each dimension.
        kernel = 1
        for size, std, mgrid in zip(
            kernel_size,
            sigma,
            torch.meshgrid([torch.arange(size, dtype=torch.float32) for size in kernel_size], indexing="ij"),
        ):  # type: ignore
            tmp_kernel = 1 / (std * math.sqrt(2 * math.pi)) * torch.exp(-(((mgrid - (size - 1) / 2) / std) ** 2) / 2)
            kernel = kernel * tmp_kernel

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())  # type: ignore
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))  # type: ignore

        self.register_buffer("weight", kernel)
        self.groups = channels

        if dim == 1:
            self.conv = F.conv1d
        elif dim == 2:
            self.conv = F.conv2d
        elif dim == 3:
            self.conv = F.conv3d
        else:
            raise RuntimeError(f"Only 1, 2 and 3 dimensions are supported. Received {dim}.")

    def forward(self, data: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`GaussianSmoothing`.

        Parameters
        ----------
        data : torch.Tensor
            Input to apply gaussian filter on.

        Returns
        -------
        torch.Tensor
            Filtered output.
        """
        if self.shift:
            data = data.permute(0, 2, 3, 1)
            data = ifft2(
                torch.fft.fftshift(
                    fft2(
                        torch.view_as_real(data[..., 0] + 1j * data[..., 1]),
                        self.fft_centered,
                        self.fft_normalization,
                        self.spatial_dims,
                    ),
                ),
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
            ).permute(0, 3, 1, 2)

        x = self.conv(data, weight=self.weight.to(data), groups=self.groups).to(data).detach()

        if self.shift:
            x = x.permute(0, 2, 3, 1)
            x = ifft2(
                torch.fft.fftshift(
                    fft2(
                        torch.view_as_real(x[..., 0] + 1j * x[..., 1]),
                        self.fft_centered,
                        self.fft_normalization,
                        self.spatial_dims,
                    ),
                ),
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
            ).permute(0, 3, 1, 2)

        return x


class LeastSquaresFitting:
    """Differentiable least square fitting in PyTorch."""

    def __init__(self, device):
        """Inits :class:`LeastSquaresFitting`."""
        super().__init__()
        self.device = device

    @staticmethod
    def lsqrt(A: torch.Tensor, Y: torch.Tensor, reg_factor: float = 0.0) -> torch.Tensor:
        """Differentiable least square solution.

        Parameters
        ----------
        A : torch.Tensor
            Input matrix.
        Y : torch.Tensor
            Echo times matrix.
        reg_factor : float
            Regularization parameter.

        Returns
        -------
        torch.Tensor
            Least square solution.
        """
        q, r = torch.qr(A)
        return torch.inverse(r) @ q.permute(0, 2, 1) @ Y + reg_factor

    @staticmethod
    def lsqrt_pinv(
        A: torch.Tensor, Y: torch.Tensor, reg_factor: float = 0.0  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Differentiable inverse least square solution.

        Parameters
        ----------
        A : torch.Tensor
            Input matrix.
        Y : torch.Tensor
            Echo times matrix.
        reg_factor : float
            Regularization parameter.

        Returns
        -------
        torch.Tensor
            Inverse least square solution.
        """
        if Y.dim() == 2:
            return torch.matmul(torch.inverse(Y), A)
        return torch.bmm(
            torch.matmul(
                torch.inverse(torch.matmul(torch.conj(Y).permute(0, 2, 1), Y)), torch.conj(Y).permute(0, 2, 1)
            ),
            A,
        )[..., 0]


def R2star_B0_S0_phi_mapping(
    prediction: torch.Tensor,
    TEs: Union[Optional[List[float]], float],
    anatomy_mask: torch.Tensor,
    scaling_factor: float = 1e-3,
    fft_centered: bool = False,
    fft_normalization: str = "backward",
    spatial_dims: Sequence[int] = None,
):
    """Maps the prediction to R2*, B0, and S0 maps.

    Parameters
    ----------
    prediction : torch.Tensor
        The prediction of the model.
    TEs : Union[Optional[List[float]], float]
        The TEs of the images.
    anatomy_mask : torch.Tensor
        The anatomy mask of the images.
    scaling_factor : float
        The scaling factor to apply to the prediction.
    fft_centered : bool
        Whether to center the FFT for a real- or complex-valued input.
    fft_normalization : str
        Whether to normalize the FFT output (None, "ortho", "backward", "forward", "none").
    spatial_dims : Sequence[int]
        Spatial dimensions to keep in the FFT.

    Returns
    -------
    R2star : torch.Tensor
        The R2* map.
    B0 : torch.Tensor
        The B0 map.
    S0 : torch.Tensor
        The S0 map.
    phi : torch.Tensor
        The phi map.
    """
    R2star_map = R2star_mapping(prediction, TEs, scaling_factor=scaling_factor)
    B0_map = -B0_phi_mapping(
        prediction,
        TEs,
        anatomy_mask,
        scaling_factor=scaling_factor,
        fft_centered=fft_centered,
        fft_normalization=fft_normalization,
        spatial_dims=spatial_dims,
    )[0]
    S0_map, phi_map = S0_mapping(
        prediction,
        TEs,
        R2star_map,
        B0_map,
        scaling_factor=scaling_factor,
        fft_centered=fft_centered,
        fft_normalization=fft_normalization,
        spatial_dims=spatial_dims,
    )

    return R2star_map, S0_map, B0_map, phi_map


def R2star_mapping(prediction: torch.Tensor, TEs: Union[Optional[List[float]], float], scaling_factor: float = 1e-3):
    """R2* map and S0 map estimation for multi-echo GRE from stored magnitude image files acquired at multiple TEs.

    Parameters
    ----------
    prediction : torch.Tensor
        The prediction of the model.
    TEs : Union[Optional[List[float]], float]
        The TEs of the images.
    scaling_factor : float
        The scaling factor.

    Returns
    -------
    R2star : torch.Tensor
        The R2* map.
    S0 : torch.Tensor
        The S0 map.
    """
    prediction = torch.view_as_complex(prediction)
    prediction = torch.abs(prediction / torch.max(torch.abs(prediction))) + 1e-8

    prediction_flatten = torch.flatten(prediction, start_dim=1, end_dim=-1).detach().cpu()
    log_prediction_flatten = torch.log(prediction_flatten)
    sqrt_prediction_flatten = torch.sqrt(prediction_flatten)

    TEs = torch.tensor(TEs).to(prediction_flatten)
    TEs = TEs * scaling_factor  # type: ignore

    R2star_map = torch.zeros([prediction_flatten.shape[1]])
    for i in range(prediction_flatten.shape[1]):
        R2star_map[i], _ = torch.from_numpy(
            np.polyfit(TEs, log_prediction_flatten[:, i], 1, w=sqrt_prediction_flatten[:, i])
        ).to(prediction)

    R2star_map = torch.reshape(-R2star_map, prediction.shape[1:4])
    return R2star_map


def B0_phi_mapping(
    prediction: torch.Tensor,
    TEs: Union[Optional[List[float]], float],
    anatomy_mask: torch.Tensor,
    scaling_factor: float = 1e-3,
    fft_centered: bool = False,
    fft_normalization: str = "backward",
    spatial_dims: Sequence[int] = None,
):
    """B0 map and Phi map estimation for multi-echo GRE from stored magnitude image files acquired at multiple TEs.

    Parameters
    ----------
    prediction : torch.Tensor
        The prediction of the model.
    TEs : Union[Optional[List[float]], float]
        The TEs of the images.
    anatomy_mask : torch.Tensor
        The anatomy mask of the images.
    scaling_factor : float
        The scaling factor.
    fft_centered : bool
        Whether to center the FFT for a real- or complex-valued input.
    fft_normalization : str
        Whether to normalize the FFT output (None, "ortho", "backward", "forward", "none").
    spatial_dims : Sequence[int]
        Spatial dimensions to keep in the FFT.

    Returns
    -------
    B0 : torch.Tensor
        The B0 map.
    phi : torch.Tensor
        The phi map.
    """
    lsq = LeastSquaresFitting(device=prediction.device)

    TEnotused = 3  # if fully_sampled else 3
    TEs = torch.tensor(TEs)

    # brain_mask is used only for descale of phase difference (so that phase_diff is in between -2pi and 2pi)
    anatomy_mask_descale = anatomy_mask
    shape = prediction.shape

    # apply gaussian blur with radius r to
    smoothing = GaussianSmoothing(
        channels=2,
        kernel_size=9,
        sigma=scaling_factor,
        dim=2,
        fft_centered=fft_centered,
        fft_normalization=fft_normalization,
        spatial_dims=spatial_dims,
    )
    prediction = prediction.unsqueeze(1).permute([0, 1, 4, 2, 3])  # add a dummy batch dimension
    for i in range(prediction.shape[0]):
        prediction[i] = smoothing(F.pad(prediction[i], (4, 4, 4, 4), mode="reflect"))
    prediction = prediction.permute([0, 1, 3, 4, 2]).squeeze(1)

    prediction = ifft2(
        torch.fft.fftshift(fft2(prediction, fft_centered, fft_normalization, spatial_dims), dim=(1, 2)),
        fft_centered,
        fft_normalization,
        spatial_dims,
    )

    phase = torch.angle(torch.view_as_complex(prediction))

    # unwrap phases
    phase_unwrapped = torch.zeros_like(phase)

    body_part_mask = anatomy_mask.clone()
    body_part_mask_np = np.invert(body_part_mask.cpu().detach().numpy() > 0.5)

    # loop over echo times
    for i in range(phase.shape[0]):
        phase_unwrapped[i] = torch.from_numpy(
            unwrap_phase(np.ma.array(phase[i].detach().cpu().numpy(), mask=body_part_mask_np)).data
        ).to(prediction)

    phase_diff_set = []
    TE_diff = []

    # obtain phase differences and TE differences
    for i in range(phase_unwrapped.shape[0] - TEnotused):
        phase_diff_set.append(torch.flatten(phase_unwrapped[i + 1] - phase_unwrapped[i]))
        phase_diff_set[i] = (
            phase_diff_set[i]
            - torch.round(
                torch.abs(
                    torch.sum(phase_diff_set[i] * torch.flatten(anatomy_mask_descale))
                    / torch.sum(anatomy_mask_descale)
                    / 2
                    / np.pi
                )
            )
            * 2
            * np.pi
        )
        TE_diff.append(TEs[i + 1] - TEs[i])  # type: ignore

    phase_diff_set = torch.stack(phase_diff_set, 0)
    TE_diff = torch.stack(TE_diff, 0).to(prediction)

    # least squares fitting to obtain phase map
    B0_map_tmp = lsq.lsqrt_pinv(
        phase_diff_set.unsqueeze(2).permute(1, 0, 2), TE_diff.unsqueeze(1) * scaling_factor  # type: ignore
    )
    B0_map = B0_map_tmp.reshape(shape[-3], shape[-2])
    B0_map = B0_map * torch.abs(body_part_mask)

    # obtain phi map
    phi_map = (phase_unwrapped[0] - scaling_factor * TEs[0] * B0_map).squeeze(0)  # type: ignore

    return B0_map.squeeze(0).to(prediction), phi_map.to(prediction)


def S0_mapping(
    prediction: torch.Tensor,
    TEs: Union[Optional[List[float]], float],
    R2star_map: torch.Tensor,
    B0_map: torch.Tensor,
    scaling_factor: float = 1e-3,
    fft_centered: bool = False,
    fft_normalization: str = "backward",
    spatial_dims: Sequence[int] = None,
):
    """Complex S0 mapping.

    Parameters
    ----------
    prediction : torch.Tensor
        The prediction of the model.
    TEs : Union[Optional[List[float]], float]
        The TEs of the images.
    R2star_map : torch.Tensor
        The R2* map.
    B0_map : torch.Tensor
        The B0 map.
    scaling_factor : float
        The scaling factor.
    fft_centered : bool
        Whether to center the FFT for a real- or complex-valued input.
    fft_normalization : str
        Whether to normalize the FFT output (None, "ortho", "backward", "forward", "none").
    spatial_dims : Sequence[int]
        Spatial dimensions to keep in the FFT.

    Returns
    -------
    S0 : torch.Tensor
        The S0 map.
    """
    lsq = LeastSquaresFitting(device=prediction.device)

    prediction = torch.view_as_complex(prediction)
    prediction_flatten = prediction.reshape(prediction.shape[0], -1)

    TEs = torch.tensor(TEs).to(prediction)

    R2star_B0_complex_map = R2star_map.to(prediction) + 1j * B0_map.to(prediction)
    R2star_B0_complex_map_flatten = R2star_B0_complex_map.flatten()

    TEs_r2 = TEs[0:4].unsqueeze(1) * -R2star_B0_complex_map_flatten  # type: ignore

    S0_map = lsq.lsqrt_pinv(
        prediction_flatten.permute(1, 0).unsqueeze(2), torch.exp(scaling_factor * TEs_r2.permute(1, 0).unsqueeze(2))
    )

    S0_map = torch.view_as_real(S0_map.reshape(prediction.shape[1:]))

    S0_map = ifft2(
        torch.fft.fftshift(fft2(S0_map, fft_centered, fft_normalization, spatial_dims), dim=(0, 1)),
        fft_centered,
        fft_normalization,
        spatial_dims,
    )

    S0_map = torch.view_as_complex(S0_map).squeeze(-1)

    return torch.abs(S0_map), torch.angle(S0_map)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from abc import ABC
from typing import Dict, Optional, Sequence, Tuple

import torch
import wandb
from omegaconf import DictConfig
from pytorch_lightning import Trainer
from torch import nn
from torch.utils.data import DataLoader
from torchmetrics.metric import Metric

from atommic.collections.common.parts import utils
from atommic.collections.common.parts.fft import ifft2
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes import modelPT
from atommic.utils import model_utils

wandb.require("service")

__all__ = ["DistributedMetricSum", "BaseMRIModel", "BaseSensitivityModel"]


# Taken and adapted from: https://github.com/facebookresearch/fastMRI/blob/main/fastmri/pl_modules/mri_module.py
class DistributedMetricSum(Metric):
    """A metric that sums the values of a metric across all workers.

    Returns
    -------
    metric : torch.FloatTensor
        The metric value.

    Examples
    --------
    >>> metric = DistributedMetricSum()
    >>> metric(torch.tensor(1.0))
    >>> metric(torch.tensor(2.0))
    >>> metric.compute()
    tensor(3.)
    """

    full_state_update: bool = True

    def __init__(self, dist_sync_on_step=True):
        """Inits :class:`DistributedMetricSum`.

        Parameters
        ----------
        dist_sync_on_step : bool
            Synchronize metric state across processes at each ``forward()`` before returning the value at the step.
            Default is ``True``.
        """
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("quantity", default=torch.tensor(0.0), dist_reduce_fx="sum")

    # pylint: disable=arguments-differ
    def update(self, batch: torch.Tensor):
        """Update the metric with a batch of data."""
        self.quantity += batch

    def compute(self):
        """Compute the metric value."""
        return self.quantity


class BaseMRIModel(modelPT.ModelPT, ABC):
    """Base class for (any task performed on) MRI models."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`BaseMRIModel`.

        Parameters
        ----------
        cfg : DictConfig
            The configuration file.
        trainer : Trainer
            The PyTorch Lightning trainer. Default is ``None``.
        """
        # Get global rank and total number of GPU workers for IterableDataset partitioning, if applicable
        self.world_size = 1
        if trainer is not None:
            self.world_size = trainer.num_nodes * trainer.num_devices

        cfg = model_utils.convert_model_config_to_dict_config(cfg)
        cfg = model_utils.maybe_update_config_version(cfg)

        super().__init__(cfg=cfg, trainer=trainer)

    # pylint: disable=arguments-differ
    def training_step(self, batch: Dict[float, torch.Tensor], batch_idx: int) -> Dict[str, torch.Tensor]:
        """Performs a training step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data.
        batch_idx : int
            Batch index.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary with the loss and the log.
        """
        raise NotImplementedError

    def validation_step(self, batch: Dict[float, torch.Tensor], batch_idx: int):
        """Performs a validation step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data.
        batch_idx : int
            Batch index.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary with the loss and the log.
        """
        raise NotImplementedError

    def test_step(self, batch: Dict[float, torch.Tensor], batch_idx: int):
        """Performs a test step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data.
        batch_idx : int
            Batch index.

        Returns
        -------
        Tuple[str, int, torch.Tensor]
            Tuple with the filename, the slice index and the prediction.
        """
        raise NotImplementedError

    def log_image(self, name, image):
        """Logs an image.

        Parameters
        ----------
        name : str
            Name of the image.
        image : torch.Tensor
            Image to log.
        """
        if image.dim() > 3:
            image = image[0, 0, :, :].unsqueeze(0)
        elif image.shape[0] != 1:
            image = image.unsqueeze(0)

        if ".h5" in name:
            name = name.replace(".h5", "")

        if "wandb" in self.logger.__module__.lower():
            if image.is_cuda:
                image = image.detach().cpu()
            self.logger.experiment.log({name: wandb.Image(image.numpy())})

        if "tensorboard" in self.logger.__module__.lower():
            self.logger.experiment.add_image(name, image, global_step=self.global_step)

    def on_validation_epoch_end(self):
        """Called at the end of validation epoch to aggregate outputs."""
        raise NotImplementedError

    def on_test_epoch_end(self):
        """Called at the end of test epoch to aggregate outputs and save predictions."""
        raise NotImplementedError

    def setup_training_data(self, train_data_config: Optional[DictConfig]):
        """Setups the training data.

        Parameters
        ----------
        train_data_config : Optional[DictConfig]
            Training data configuration.

        Returns
        -------
        train_data : torch.utils.data.DataLoader
            Training data.
        """
        self._train_dl = self._setup_dataloader_from_config(cfg=train_data_config)

    def setup_validation_data(self, val_data_config: Optional[DictConfig]):
        """Setups the validation data.

        Parameters
        ----------
        val_data_config : Optional[DictConfig]
            Validation data configuration.

        Returns
        -------
        val_data : torch.utils.data.DataLoader
            Validation data.
        """
        self._validation_dl = self._setup_dataloader_from_config(cfg=val_data_config)

    def setup_test_data(self, test_data_config: Optional[DictConfig]):
        """Setups the test data.

        Parameters
        ----------
        test_data_config : Optional[DictConfig]
            Test data configuration.

        Returns
        -------
        test_data : torch.utils.data.DataLoader
            Test data.
        """
        self._test_dl = self._setup_dataloader_from_config(cfg=test_data_config)

    @staticmethod
    def _setup_dataloader_from_config(cfg: DictConfig) -> DataLoader:
        """Setups the dataloader from the configuration (yaml) file.

        Parameters
        ----------
        cfg : DictConfig
            Configuration file.

        Returns
        -------
        dataloader : torch.utils.data.DataLoader
            Dataloader.
        """
        raise NotImplementedError


class BaseSensitivityModel(nn.Module, ABC):
    """Model for learning sensitivity estimation from k-space data [1]_. This model applies an IFFT to multichannel
    k-space data and then a U-Net to the coil images to estimate coil sensitivities.

    References
    ----------
    .. [1] Sriram A, Zbontar J, Murrell T, Defazio A, Zitnick CL, Yakubova N, Knoll F, Johnson P. End-to-end
        variational networks for accelerated MRI reconstruction. InInternational Conference on Medical Image Computing
        and Computer-Assisted Intervention 2020 Oct 4 (pp. 64-73). Springer, Cham.

    Returns
    -------
    torch.Tensor
        Estimated coil sensitivity maps.
    """

    def __init__(
        self,
        chans: int = 8,
        num_pools: int = 4,
        in_chans: int = 2,
        out_chans: int = 2,
        drop_prob: float = 0.0,
        padding_size: int = 15,
        mask_type: str = "2D",
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 1,
        normalize: bool = True,
        mask_center: bool = True,
    ):
        """Inits :class:`BaseSensitivityModel`.

        Parameters
        ----------
        chans : int
            Number of channels in the input k-space data. Default is ``8``.
        num_pools : int
            Number of U-Net downsampling/upsampling operations. Default is ``4``.
        in_chans : int
            Number of input channels to the U-Net. Default is ``2``.
        out_chans : int
            Number of output channels to the U-Net. Default is ``2``.
        drop_prob : float
            Dropout probability. Default is ``0.0``.
        padding_size : int
            Padding size for the U-Net. Default is ``15``.
        mask_type : str
            If kspace is undersampled, the undersampling mask type must be specified. Default is ``"2D"``.
        fft_centered : bool
            If ``True``, the input data is assumed to be centered. Default is ``False``.
        fft_normalization : str
            Normalization for the IFFT. Default is ``"backward"``.
        spatial_dims : Sequence[int]
            Spatial dimensions of the input data. Default is ``None``.
        coil_dim : int
            Dimension of the coils. Default is ``1``.
        normalize : bool
            If ``True``, the input data is normalized. Default is ``True``.
        mask_center : bool
            If ``True``, the center of the k-space is masked out. Default is ``True``.
        """
        super().__init__()

        self.mask_type = mask_type

        self.norm_unet = NormUnet(
            chans,
            num_pools,
            in_chans=in_chans,
            out_chans=out_chans,
            drop_prob=drop_prob,
            padding_size=padding_size,
            normalize=normalize,
        )

        self.mask_center = mask_center
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim
        self.normalize = normalize

    @staticmethod
    def chans_to_batch_dim(x: torch.Tensor) -> Tuple[torch.Tensor, int]:
        """Combines the channel dimension with the batch dimension.

        Parameters
        ----------
        x : torch.Tensor
            Tensor to stack the batch and channel dimensions.

        Returns
        -------
        Tuple[torch.Tensor, int]
            Tuple of the converted tensor and the batch size.
        """
        batch_size, coils, height, width, complex_dim = x.shape
        return x.view(batch_size * coils, 1, height, width, complex_dim), batch_size

    @staticmethod
    def batch_chans_to_chan_dim(x: torch.Tensor, batch_size: int) -> torch.Tensor:
        """Splits the batch and channel dimensions into the channel dimension.

        Parameters
        ----------
        x : torch.Tensor
            Tensor to split the batch and channel dimensions.
        batch_size : int
            Batch size.

        Returns
        -------
        torch.Tensor
            Converted tensor.
        """
        batch_size_coils, _, height, width, complex_dim = x.shape
        coils = torch.div(batch_size_coils, batch_size, rounding_mode="trunc")
        return x.view(batch_size, coils, height, width, complex_dim)

    @staticmethod
    def divide_root_sum_of_squares(x: torch.Tensor, coil_dim: int) -> torch.Tensor:
        """Divide the input by the root of the sum of squares of the magnitude of each complex number.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor to divide.
        coil_dim : int
            Dimension of the coils.

        Returns
        -------
        torch.Tensor
            Normalized tensor by the root sum of squares.
        """
        return x / utils.rss_complex(x, dim=coil_dim).unsqueeze(-1).unsqueeze(coil_dim)

    @staticmethod
    def get_pad_and_num_low_freqs(
        mask: torch.Tensor, num_low_frequencies: Optional[int] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Get the padding to apply to the input to make it square and the number of low frequencies to keep.

        Parameters
        ----------
        mask : torch.Tensor
            Mask to use to determine the padding and number of low frequencies.
        num_low_frequencies : Optional[int]
            Number of low frequencies to keep. Default is ``None``.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            Tuple of the padding to apply to the input and the number of low frequencies to keep.
        """
        if num_low_frequencies is None or num_low_frequencies == 0:
            # get low frequency line locations and mask them out
            squeezed_mask = mask[:, 0, 0, :, 0].to(torch.int8)
            cent = torch.div(squeezed_mask.shape[1], 2, rounding_mode="trunc")
            # running argmin returns the first non-zero
            left = torch.argmin(squeezed_mask[:, :cent].flip(1), dim=1)
            right = torch.argmin(squeezed_mask[:, cent:], dim=1)
            num_low_frequencies_tensor = torch.max(
                2 * torch.min(left, right), torch.ones_like(left)
            )  # force a symmetric center unless 1
        else:
            num_low_frequencies_tensor = num_low_frequencies * torch.ones(
                mask.shape[0], dtype=mask.dtype, device=mask.device
            )

        pad = torch.div(mask.shape[-2] - num_low_frequencies_tensor + 1, 2, rounding_mode="trunc")

        return pad, num_low_frequencies_tensor

    def __call__(
        self,
        masked_kspace: torch.Tensor,
        mask: torch.Tensor,
        sensitivity_maps: Optional[torch.Tensor] = None,
        num_low_frequencies: Optional[int] = None,
    ) -> torch.Tensor:
        """Calls :class:`BaseSensitivityModel`.

        Parameters
        ----------
        masked_kspace : torch.Tensor
            Subsampled k-space data of shape [batch_size, n_coils, n_x, n_y, 2].
        mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        sensitivity_maps : Optional[torch.Tensor]
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. Default is ``None``. If provided, the
            sensitivity maps will be fine-tuned using a U-Net.
        num_low_frequencies : Optional[int]
            Number of low frequencies to keep. Default is ``None``.

        Returns
        -------
        torch.Tensor
            Estimated coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        """
        return self.forward(masked_kspace, mask, sensitivity_maps, num_low_frequencies)

    def forward(
        self,
        masked_kspace: torch.Tensor,
        mask: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        num_low_frequencies: Optional[int] = None,
    ) -> torch.Tensor:
        """Forward pass of :class:`BaseSensitivityModel`.

        Parameters
        ----------
        masked_kspace : torch.Tensor
            Subsampled k-space data of shape [batch_size, n_coils, n_x, n_y, 2].
        mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. If provided, the sensitivity maps will
            be fine-tuned using a U-Net. If not provided, the sensitivity maps will be estimated from the input k-space
            data. In the second case the sensitivity maps are just a torch.ones_like of the input k-space data, so the
            mean of the sensitivity maps is 1. This is the default behavior of the model.
        num_low_frequencies : Optional[int]
            Number of low frequencies to keep. Default is ``None``.

        Returns
        -------
        torch.Tensor
            Estimated coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        """
        if torch.mean(sensitivity_maps) != 1:
            # If sensitivity maps are provided, fine-tune them using a U-Net.
            initial_sensitivity_maps = sensitivity_maps.clone()
        else:
            if self.mask_center:
                # get low frequency line locations and mask them out in the center of k-space
                pad, num_low_freqs = self.get_pad_and_num_low_freqs(mask, num_low_frequencies)
                masked_kspace = utils.batched_mask_center(
                    masked_kspace, pad, pad + num_low_freqs, mask_type=self.mask_type
                )

            initial_sensitivity_maps = ifft2(
                masked_kspace,
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )

        # stack batch and coil dimensions
        initial_sensitivity_maps, batches = self.chans_to_batch_dim(initial_sensitivity_maps)

        # estimate sensitivities
        sensitivity_maps = self.batch_chans_to_chan_dim(self.norm_unet(initial_sensitivity_maps), batches)

        # normalize
        if self.normalize:
            sensitivity_maps = self.divide_root_sum_of_squares(sensitivity_maps, self.coil_dim)

        return sensitivity_maps



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Optional, Tuple

import torch

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul


class Conv2d(torch.nn.Module):
    """Implementation of a simple cascade of 2D convolutions. If batchnorm is set to True, batch normalization layer is
    applied after each convolution.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int,
        n_convs: int = 3,
        activation: torch.nn.Module = torch.nn.PReLU(),
        batchnorm: bool = False,
    ):
        """Inits :class:`Conv2d`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        hidden_channels : int
            Number of hidden channels.
        n_convs : int, optional
            Number of convolutional layers. Default is ``3``.
        activation : torch.nn.Module, optional
            Activation function. Default is ``nn.PReLU()``.
        batchnorm : bool, optional
            If True a batch normalization layer is applied after every convolution. Default is ``False``.
        """
        super().__init__()

        self.conv = []
        for idx in range(n_convs):
            self.conv.append(
                torch.nn.Conv2d(
                    in_channels if idx == 0 else hidden_channels,
                    hidden_channels if idx != n_convs - 1 else out_channels,
                    kernel_size=3,
                    padding=1,
                )
            )
            if batchnorm:
                self.conv.append(
                    torch.nn.BatchNorm2d(hidden_channels if idx != n_convs - 1 else out_channels, eps=1e-4)
                )
            if idx != n_convs - 1:
                self.conv.append(activation)
        self.conv = torch.nn.Sequential(*self.conv)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`Conv2d`."""
        if x.dim() == 5:
            x = x.squeeze(1)
            if x.shape[-1] == 2:
                x = x.permute(0, 3, 1, 2)
        return self.conv(x)  # type: ignore


class CascadeNetBlock(torch.nn.Module):
    """Model block for CascadeNet & Convolution Recurrent Neural Network."""

    def __init__(
        self,
        model: torch.nn.Module,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        no_dc: bool = False,
    ):
        """Inits :class:`CascadeNetBlock`.

        Parameters
        ----------
        model : torch.nn.Module
            Model to apply soft data consistency.
        fft_centered : bool, optional
            Whether to center the FFT. Default is ``False``.
        fft_normalization : str, optional
            Whether to normalize the FFT. Default is ``"backward"``.
        spatial_dims : Tuple[int, int], optional
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        no_dc : bool, optional
            Flag to disable the soft data consistency. Default is ``False``.
        """
        super().__init__()

        self.model = model
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim
        self.no_dc = no_dc
        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    def sens_expand(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with coil-combined data to get multicoil data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            Expanded multicoil data.
        """
        return fft2(
            complex_mul(x, sens_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )

    def sens_reduce(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with multicoil data to get coil-combined data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            SENSE coil-combined reconstruction.
        """
        x = ifft2(x, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
        return complex_mul(x, complex_conj(sens_maps)).sum(dim=self.coil_dim)

    def forward(
        self,
        pred: torch.Tensor,
        ref_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of :class:`CascadeNetBlock`.

        Parameters
        ----------
        pred : torch.Tensor
            Predicted k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        ref_kspace : torch.Tensor
            Reference k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]

        Returns
        -------
        torch.Tensor
            Reconstructed image. Shape [batch_size, n_x, n_y, 2]
        """
        zero = torch.zeros(1, 1, 1, 1, 1).to(pred)
        soft_dc = torch.where(mask.bool(), pred - ref_kspace, zero) * self.dc_weight

        prediction = self.sens_reduce(pred, sensitivity_maps)
        prediction = self.model(prediction.squeeze(self.coil_dim).permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
        if prediction.dim() < sensitivity_maps.dim():
            prediction = prediction.unsqueeze(1)
        prediction = self.sens_expand(prediction, sensitivity_maps)

        if not self.no_dc:
            prediction = pred - soft_dc - prediction

        return prediction



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.segmentation.nn.lambdaunet_base.lambdaunet_block import LambdaUNet
from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet
from atommic.core.classes.common import typecheck

__all__ = ["SegmentationLambdaUNet"]


class SegmentationLambdaUNet(BaseSegmentationNet):
    """Implementation of the Lambda UNet for MRI segmentation, as presented in [Yanglan2021]_.

    References
    ----------
    .. [Yanglan2021] Yanglan Ou, Ye Yuan, Xiaolei Huang, Kelvin Wong, John Volpi, James Z. Wang, Stephen T.C. Wong.
        LambdaUNet: 2.5D Stroke Lesion Segmentation of Diffusion-weighted MR Images. 2021.
        https://arxiv.org/abs/2104.13917

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        return LambdaUNet(
            in_chans=self.input_channels,
            out_chans=cfg.get("segmentation_module_output_channels", 2),
            chans=cfg.get("segmentation_module_channels", 32),
            num_pool_layers=cfg.get("segmentation_module_pooling_layers", 4),
            drop_prob=cfg.get("segmentation_module_dropout", 0.0),
            query_depth=cfg.get("segmentation_module_query_depth", 16),
            intra_depth=cfg.get("segmentation_module_intra_depth", 4),
            receptive_kernel=cfg.get("segmentation_module_receptive_kernel", 3),
            temporal_kernel=cfg.get("segmentation_module_temporal_kernel", 1),
            num_slices=self.consecutive_slices,
        )

    @typecheck()
    def forward(self, image: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Forward pass of the network.

        Parameters
        ----------
        image : torch.Tensor
            Input image. Shape [batch_size, n_x, n_y] or [batch_size, n_x, n_y, 2]
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Predicted segmentation. Shape [batch_size, n_classes, n_x, n_y]
        """
        # Adjust the dimensions of the input image
        if image.shape[-1] == 2:
            if self.input_channels == 1:
                image = torch.view_as_complex(image).unsqueeze(1)
                if self.magnitude_input:
                    image = torch.abs(image)
            elif self.input_channels == 2 and not self.magnitude_input:
                image = image.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")

        mean = 1.0
        std = 1.0
        if self.normalize:
            image, mean, std = self.norm(image)
        image, pad_sizes = self.pad(image)
        if self.consecutive_slices > 1:
            batch, slices = image.shape[:2]
            image = image.reshape(batch * slices, *image.shape[2:])
        segmentation = self.segmentation_module(image)
        segmentation = self.unpad(segmentation, *pad_sizes)
        if self.normalize:
            segmentation = self.unnorm(segmentation, mean, std)

        if self.normalize_segmentation_output:
            segmentation = (segmentation - segmentation.min()) / (segmentation.max() - segmentation.min())

        if self.consecutive_slices > 1:
            segmentation = segmentation.reshape(batch, slices, *segmentation.shape[1:])

        return torch.abs(segmentation)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/lightning_logger_patch.py

import logging as _logging
from logging.handlers import MemoryHandler
from typing import Any, Dict

import pytorch_lightning as pl

HANDLERS: Dict[Any, Any] = {}


def add_memory_handlers_to_pl_logger():
    """Adds two MemoryHandlers to pytorch_lightning's logger. These two handlers are essentially message buffers. This
    function is called in atommic.utils.__init__.py. These handlers are used in add_filehandlers_to_pl_logger to flush
    buffered messages to files.
    """
    if not HANDLERS:
        HANDLERS["memory_err"] = MemoryHandler(-1)
        HANDLERS["memory_err"].addFilter(lambda record: record.levelno > _logging.INFO)
        HANDLERS["memory_all"] = MemoryHandler(-1)
        pl._logger.addHandler(HANDLERS["memory_err"])  # pylint: disable=protected-access
        pl._logger.addHandler(HANDLERS["memory_all"])  # pylint: disable=protected-access


def add_filehandlers_to_pl_logger(all_log_file, err_log_file):
    """Adds two filehandlers to pytorch_lightning's logger. Called in atommic.utils.exp_manager(). The first
    filehandler logs all messages to all_log_file while the second filehandler logs all WARNING and higher messages to
    err_log_file. If "memory_err" and "memory_all" exist in HANDLERS, then those buffers are flushed to err_log_file
    and all_log_file respectively, and then closed.
    """
    HANDLERS["file"] = _logging.FileHandler(all_log_file)
    pl._logger.addHandler(HANDLERS["file"])  # pylint: disable=protected-access
    HANDLERS["file_err"] = _logging.FileHandler(err_log_file)
    HANDLERS["file_err"].addFilter(lambda record: record.levelno > _logging.INFO)
    pl._logger.addHandler(HANDLERS["file_err"])  # pylint: disable=protected-access

    if HANDLERS.get("memory_all"):
        HANDLERS["memory_all"].setTarget(HANDLERS["file"])
        HANDLERS["memory_all"].close()
        del HANDLERS["memory_all"]
    if HANDLERS.get("memory_err"):
        HANDLERS["memory_err"].setTarget(HANDLERS["file_err"])
        HANDLERS["memory_err"].close()
        del HANDLERS["memory_err"]
